from __future__ import annotations

from typing import TYPE_CHECKING

from angr.analyses.decompiler import DECOMPILATION_PRESETS
from angr.analyses.decompiler.decompilation_options import options as dec_options
from angr.analyses.decompiler.optimization_passes import get_optimization_passes
from angr.analyses.decompiler.peephole_optimizations import EXPR_OPTS, MULTI_STMT_OPTS, STMT_OPTS
from PySide6.QtCore import QSize
from PySide6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTreeWidget,
    QVBoxLayout,
    QWidget,
)

from angrmanagement.ui.widgets.qproperty_editor import (
    BoolPropertyItem,
    ComboPropertyItem,
    GroupPropertyItem,
    PropertyModel,
    QPropertyEditor,
)

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.views.code_view import CodeView


def map_option_to_property(option, enabled):
    if hasattr(option, "value_type") and not isinstance(option.value_type, bool) and option.candidate_values:
        return ComboPropertyItem(
            option.NAME, option.default_value, option.candidate_values, description=option.DESCRIPTION, option=option
        )
    else:
        return BoolPropertyItem(option.NAME, enabled, description=option.DESCRIPTION, option=option)


class QDecompilationOptions(QWidget):
    """
    The widget for selecting values for decompilation options. Will synchronize its status back to its parent (passed
    in parameter) code view with the Apply button is pressed.

    Since some options have default values depending on the current arch and os, it is important to call
    reload(force=True) to reset values to their defaults whenever the current project changes.
    """

    def __init__(self, code_view, instance: Instance) -> None:
        super().__init__()

        self.dirty = True

        self._code_view: CodeView = code_view
        self._instance = instance
        self._options = None
        self._opti_passes = None
        self._peephole_opts = None

        # widgets
        self._preset_cmb: QComboBox
        self._tree_view: QTreeWidget
        self._apply_btn: QPushButton

        self._qoptions = []
        self._qoptipasses = []
        self._qpeephole_opts = []

        self._init_widgets()

        self.reload(True)

    def sizeHint(self):  # pylint:disable=no-self-use
        return QSize(400, 400)

    def reload(self, force: bool = False) -> None:
        if force or self._options is None:
            self._options = self.get_default_options()

        if force or self._opti_passes is None:
            if not self._instance.project.am_none:
                self._opti_passes = self.get_all_passes()
            else:
                self._opti_passes = []

        if force or self._peephole_opts is None:
            self._peephole_opts = self.get_all_peephole_opts()

        self._reload_options(force)

    def _on_item_changed(self, item, _) -> None:
        if getattr(item.extra["option"], "clears_cache", True):
            self.dirty = True

    def _on_apply_pressed(self) -> None:
        if self.dirty:
            self.dirty = False
            # clear the cached version
            self._code_view.decompile(reset_cache=True)
        else:
            if not self._code_view.codegen.am_none:
                self._code_view.codegen.reapply_options(self.option_and_values)
                self._code_view.codegen.am_event()

    @property
    def selected_passes(self):
        selected = []
        for item in self._qoptipasses:
            if item.value:
                selected.append(item.extra["option"])
        return selected

    @property
    def selected_peephole_opts(self):
        selected = []
        for item in self._qpeephole_opts:
            if item.value:
                selected.append(item.extra["option"])
        return selected

    @property
    def option_and_values(self):
        return [(item.extra["option"], item.value) for item in self._qoptions]

    def get_default_options(self):  # pylint: disable=no-self-use
        return dec_options

    def get_default_passes(self):
        if self._instance is None or self._instance.project.am_none:
            return []
        default_preset = DECOMPILATION_PRESETS[self._preset_cmb.currentText()]
        return default_preset.get_optimization_passes(
            self._instance.project.arch, self._instance.project.simos.name
        ) + [x for x, de in self._code_view.workspace.plugins.optimization_passes() if de]

    def get_all_passes(self):
        if self._instance is None or self._instance.project.am_none:
            return []
        return get_optimization_passes(self._instance.project.arch, self._instance.project.simos.name) + [
            x for x, _ in self._code_view.workspace.plugins.optimization_passes()
        ]

    def get_default_peephole_opts(self):  # pylint: disable=no-self-use
        return MULTI_STMT_OPTS + STMT_OPTS + EXPR_OPTS

    def get_all_peephole_opts(self):  # pylint: disable=no-self-use
        return MULTI_STMT_OPTS + STMT_OPTS + EXPR_OPTS

    def _init_widgets(self) -> None:
        preset_lyt = QHBoxLayout()
        preset_lyt.setContentsMargins(0, 0, 0, 0)
        preset_lyt.setSpacing(3)

        preset_lyt.addWidget(QLabel("Preset:"))
        self._preset_cmb = QComboBox()
        presets = sorted([n for n in DECOMPILATION_PRESETS if n != "default"])
        self._preset_cmb.addItems(presets)
        self._preset_cmb.setCurrentIndex(
            next(i for i, n in enumerate(presets) if DECOMPILATION_PRESETS[n] is DECOMPILATION_PRESETS["default"])
        )
        self._preset_cmb.activated.connect(lambda: self.reload(force=True))
        preset_lyt.addWidget(self._preset_cmb, 1)

        self._tree_view = QPropertyEditor()

        # refresh button
        self._apply_btn = QPushButton("Apply")
        self._apply_btn.clicked.connect(self._on_apply_pressed)

        layout = QVBoxLayout()
        layout.setContentsMargins(3, 3, 3, 3)
        layout.setSpacing(3)
        layout.addLayout(preset_lyt)
        layout.addWidget(self._tree_view)
        layout.addWidget(self._apply_btn)

        self.setLayout(layout)

    def _reload_options(self, reset_values: bool = False) -> None:
        vals_options = dict(self.option_and_values)
        vals_peephole = self.selected_peephole_opts
        vals_passes = self.selected_passes

        self._qoptions.clear()
        self._qoptipasses.clear()
        self._qpeephole_opts.clear()

        categories = {}
        root = GroupPropertyItem("Root")
        model = PropertyModel(root)
        model.valueChanged.connect(self._on_item_changed)

        # populate the tree widget with new options
        for option in sorted(self._options, key=lambda x: x.NAME):
            if option.category in categories:
                category = categories[option.category]
            else:
                category = GroupPropertyItem(option.category, description=option.category)
                categories[option.category] = category
                root.addChild(category)

            enabled = option.default_value if reset_values else vals_options.get(option, option.default_value)
            w = map_option_to_property(option, enabled)
            category.addChild(w)
            self._qoptions.append(w)

        passes_category = GroupPropertyItem("Optimization Passes", description="Optimization Passes")
        categories["passes"] = passes_category
        root.addChild(passes_category)

        default_passes = set(self.get_default_passes())
        for pass_ in self._opti_passes:
            enabled = pass_ in default_passes if reset_values else pass_ in vals_passes
            w = map_option_to_property(pass_, enabled)
            passes_category.addChild(w)
            self._qoptipasses.append(w)

        po_category = GroupPropertyItem("Peephole Optimizations", description="Peephole Optimizations")
        categories["peephole_opts"] = po_category
        root.addChild(po_category)

        default_peephole_opts = self.get_default_peephole_opts()
        for opt_ in self._peephole_opts:
            enabled = opt_ in default_peephole_opts if reset_values else opt_ in vals_peephole
            w = map_option_to_property(opt_, enabled)
            po_category.addChild(w)
            self._qpeephole_opts.append(w)

        self._tree_view.setModel(model)
        self.dirty = True



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFrame,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QPushButton,
    QTabWidget,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
)

from angrmanagement.data.jobs import SimgrExploreJob, SimgrStepJob
from angrmanagement.logic.threads import gui_thread_schedule
from angrmanagement.ui.widgets.qsimulation_manager_viewer import QSimulationManagerViewer

if TYPE_CHECKING:
    from angr import SimState

    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class QSimulationManagers(QFrame):
    def __init__(self, workspace: Workspace, instance: Instance, simgr, state, parent=None) -> None:
        """
        :param instance:                The data source for this project
        :param object parent:           The parent widget.
        """
        super().__init__(parent)

        self.workspace = workspace
        self.instance = instance
        self.simgrs = instance.simgrs
        self.simgr = simgr
        self.state = state

        self._simgrs_list: QComboBox
        self._avoids_list: QTreeWidget
        self._finds_list: QTreeWidget
        self._simgr_viewer: QSimulationManagerViewer
        self._oneactive_checkbox: QCheckBox

        self._init_widgets()
        self.refresh()

        self.simgr.am_subscribe(self._watch_simgr)
        self.simgrs.am_subscribe(self._watch_simgrs)
        self.state.am_subscribe(self._watch_state)

    @property
    def find_addrs(self):
        return list({int(item.text(0), 16) for item in self._get_checked_items(self._finds_list)})

    @property
    def avoid_addrs(self):
        return list({int(item.text(0), 16) for item in self._get_checked_items(self._avoids_list)})

    def closeEvent(self, _) -> None:
        self.simgr.am_unsubscribe(self._watch_simgr)
        self.simgrs.am_unsubscribe(self._watch_simgrs)
        self.state.am_unsubscribe(self._watch_state)

    #
    # Public methods
    #

    def refresh(self) -> None:
        self._simgrs_list.clear()
        for i, simgr in enumerate(self.simgrs):
            self._simgrs_list.addItem(simgr.am_name)
            if simgr is self.simgr.am_obj:
                self._simgrs_list.setCurrentIndex(i)

    def add_avoid_address(self, addr: int) -> None:
        self.add_address_to_list(self._avoids_list, addr)

    def add_find_address(self, addr: int) -> None:
        self.add_address_to_list(self._finds_list, addr)

    def remove_find_address(self, addr: int) -> None:
        self._remove_addr(self._finds_list, addr)

    def remove_avoid_address(self, addr: int) -> None:
        self._remove_addr(self._avoids_list, addr)

    @staticmethod
    def add_address_to_list(qtreelist: QTreeWidget, addr: int):
        for i in range(qtreelist.topLevelItemCount()):
            item: QTreeWidgetItem = qtreelist.topLevelItem(i)
            if int(item.text(0), 16) == addr:
                return None  # deduplicate

        item = QTreeWidgetItem(qtreelist)
        item.setText(0, f"{addr:#x}")
        item.setFlags(item.flags() | Qt.ItemFlag.ItemIsUserCheckable)
        item.setData(0, Qt.ItemDataRole.CheckStateRole, Qt.CheckState.Checked)
        return item

    #
    # Initialization
    #

    def _init_widgets(self) -> None:
        tab = QTabWidget()

        self._init_simgrs_tab(tab)
        self._init_settings_tab(tab)
        self._init_avoids_tab(tab)
        self._init_finds_tab(tab)

        layout = QVBoxLayout()
        layout.addWidget(tab)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def select_states(self, states: list[SimState]) -> None:
        stash_tree_item = self._simgr_viewer.get_stash_tree_item("active")
        states_set = set(states)
        for state_tree_item in stash_tree_item:
            if state_tree_item.state in states_set:
                state_tree_item.setSelected(True)
        stash_tree_item.setExpanded(True)

    def _init_simgrs_tab(self, tab) -> None:
        # simgrs list

        simgrs_label = QLabel(self)
        simgrs_label.setText("Simulation Manager")

        simgrs_list = QComboBox(self)
        self._simgrs_list = simgrs_list
        simgrs_list.currentIndexChanged.connect(self._on_simgr_selection)

        pg_layout = QHBoxLayout()
        pg_layout.addWidget(simgrs_label)
        pg_layout.addWidget(simgrs_list)

        # simulation manager information
        viewer = QSimulationManagerViewer(self.simgr)
        self._simgr_viewer = viewer
        viewer.currentItemChanged.connect(self._on_state_selection)

        #
        # Buttons
        #

        # step button
        step_button = QPushButton()
        step_button.setText("Step actives")
        step_button.released.connect(self._on_step_clicked)

        # step until branch
        step_until_branch_button = QPushButton("Step actives until branch")
        step_until_branch_button.released.connect(self._on_step_until_branch_clicked)

        # explore button
        explore_button = QPushButton("Explore")
        explore_button.released.connect(self._on_explore_clicked)

        # buttons layout
        buttons_layout = QVBoxLayout()
        layout = QHBoxLayout()
        layout.addWidget(explore_button)
        buttons_layout.addLayout(layout)

        layout = QHBoxLayout()
        layout.addWidget(step_button)
        layout.addWidget(step_until_branch_button)
        buttons_layout.addLayout(layout)

        simgrs_layout = QVBoxLayout()
        simgrs_layout.addLayout(pg_layout)
        simgrs_layout.addWidget(viewer)
        simgrs_layout.addLayout(buttons_layout)

        frame = QFrame()
        frame.setLayout(simgrs_layout)

        tab.addTab(frame, "General")

    def _init_settings_tab(self, tab) -> None:
        oneactive_checkbox = QCheckBox("Keep at most one active path")
        oneactive_checkbox.setChecked(False)
        self._oneactive_checkbox = oneactive_checkbox

        settings_layout = QVBoxLayout()
        settings_layout.addWidget(oneactive_checkbox)
        settings_layout.addStretch(0)

        frame = QFrame()
        frame.setLayout(settings_layout)

        tab.addTab(frame, "Settings")

    def _init_avoids_tab(self, tab) -> None:
        avoids_list = QTreeWidget()
        avoids_list.setHeaderHidden(True)
        self._avoids_list = avoids_list

        layout = QVBoxLayout()
        layout.addWidget(avoids_list)

        import_button = QPushButton("Import List")
        import_button.clicked.connect(lambda: self._import_to_list(self._avoids_list))
        layout.addWidget(import_button)

        frame = QFrame()
        frame.setLayout(layout)

        tab.addTab(frame, "Avoids")

        self._avoids_list.itemChanged.connect(self._on_explore_addr_changed)

    def _init_finds_tab(self, tab) -> None:
        finds_list = QTreeWidget()
        finds_list.setHeaderHidden(True)
        self._finds_list = finds_list

        layout = QVBoxLayout()
        layout.addWidget(finds_list)

        frame = QFrame()
        frame.setLayout(layout)

        import_button = QPushButton("Import List")
        import_button.clicked.connect(lambda: self._import_to_list(self._finds_list))
        layout.addWidget(import_button)

        tab.addTab(frame, "Finds")

        self._finds_list.itemChanged.connect(self._on_explore_addr_changed)

    #
    # Event handlers
    #

    def _on_step_clicked(self) -> None:
        if not self.simgr.am_none:
            self.workspace.job_manager.add_job(
                SimgrStepJob(
                    self.instance,
                    self.simgr.am_obj,
                    until_branch=False,
                    step_callback=self.workspace.plugins.step_callback,
                )
            )

    def _on_step_until_branch_clicked(self) -> None:
        if not self.simgr.am_none:
            self.workspace.job_manager.add_job(
                SimgrStepJob(
                    self.instance,
                    self.simgr.am_obj,
                    until_branch=True,
                    step_callback=self.workspace.plugins.step_callback,
                )
            )

    def _on_explore_clicked(self) -> None:
        if not self.simgr.am_none:

            def _step_callback(simgr):
                self.workspace.plugins.step_callback(simgr)
                if self._oneactive_checkbox.isChecked():
                    self._filter_actives(simgr, events=False)
                gui_thread_schedule(lambda: self.simgr.am_event(src="post_step"))
                return simgr

            self.workspace.job_manager.add_job(
                SimgrExploreJob.create(
                    self.instance,
                    self.simgr,
                    avoid=self.avoid_addrs,
                    find=self.find_addrs,
                    step_callback=_step_callback,
                )
            )

    def _on_simgr_selection(self) -> None:
        i = self._simgrs_list.currentIndex()
        simgr = self.simgrs[i] if i != -1 else None

        if simgr != self.simgr.am_obj:
            self.simgr.am_obj = simgr
            self.simgr.am_event(src="clicked")

    def _on_state_selection(self) -> None:
        state = self._simgr_viewer.current_state()
        if state != self.state:
            self.state.am_obj = state
            self.state.am_event(src="clicked")

    def _watch_simgr(self, **kwargs) -> None:
        if kwargs.get("src") in ("clicked", "filter_actives", "post_step"):
            return
        elif kwargs.get("src") == "job_done" and kwargs.get("job") == "step":
            self._filter_actives(self.simgr)
        else:
            idx = self._simgrs_list.findText(self.simgr.am_obj.am_name)
            self._simgrs_list.setCurrentIndex(idx)

    def _watch_state(self, **kwargs) -> None:
        if kwargs.get("src") == "clicked":
            return

        self._simgr_viewer.select_state(self.state.am_obj)

    def _watch_simgrs(self, **_) -> None:
        self.refresh()

    def _on_explore_addr_changed(self, item: QTreeWidgetItem) -> None:  # pylint: disable=unused-argument
        """Refresh the disassembly view when an address in the 'avoids' or 'finds' tab is toggled. Ensures that
        annotations next to instructions are updated."""
        view_manager = self.workspace.view_manager
        if len(view_manager.views_by_category["disassembly"]) == 1:
            view = view_manager.first_view_in_category("disassembly")
        else:
            view = view_manager.current_view_in_category("disassembly")
        if view is not None:
            view.refresh()

    #
    # Private methods
    #

    def _filter_actives(self, simgr, events: bool = True) -> bool:
        if not self._oneactive_checkbox.isChecked():
            return False
        if len(simgr.active) < 2:
            return False

        stashed = simgr.active[1:]
        simgr.stashes["stashed"].extend(stashed)
        simgr.stashes["active"] = simgr.active[:1]
        if events:
            simgr.am_event(src="filter_actives", filtered=stashed)
        return True

    @staticmethod
    def _get_checked_items(qlist: QTreeWidget):
        items = []
        for i in range(qlist.topLevelItemCount()):
            item = qlist.topLevelItem(i)
            if item.checkState(0) == Qt.CheckState.Checked:
                items.append(item)
                for j in range(item.childCount()):
                    sub_item = item.child(j)
                    if sub_item.checkState(0) == Qt.CheckState.Checked:
                        items.append(sub_item)
        return items

    def _import_to_list(self, qlist: QTreeWidget) -> None:
        text, ok = QInputDialog.getMultiLineText(
            self, "Input Address List", "Address in hex (one address each line):", ""
        )
        if not ok:
            return
        for line in text.splitlines(keepends=False):
            try:
                addr = int(line, 16)
                self.add_address_to_list(qlist, addr)
            except ValueError:  # pylint: disable=unused-variable
                pass

    @staticmethod
    def _remove_addr(qlist: QTreeWidget, addr: int) -> None:
        for i in range(qlist.topLevelItemCount()):
            qitem = qlist.topLevelItem(i)
            if int(qitem.text(0), 16) == addr:
                qlist.takeTopLevelItem(i)
                return



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from angr.analyses.data_dep import ConstantDepNode, TmpDepNode
from PySide6 import QtCore, QtGui, QtWidgets

from angrmanagement.config import Conf

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angr.analyses.data_dep import BaseDepNode
    from capstone import CsInsn

    from angrmanagement.ui.views.data_dep_view import DataDepView

_l = logging.getLogger(__name__)


class QDataDepGraphBlock(QCachedGraphicsItem):
    """Represents a node in a QDataDepGraph, used to visualize a BaseDepNode"""

    HORIZONTAL_PADDING = 20
    VERTICAL_PADDING = 20
    LINE_MARGIN = 3

    DEFAULT_BACKGROUND = QtGui.QColor(0xB5, 0xF8, 0xFE)  # Blue
    CONSTANT_BACKGROUND = QtGui.QColor(0x87, 0xFF, 0x65)  # Green
    TMP_BACKGROUND = QtGui.QColor(0xEF, 0x47, 0x6F)  # Pink

    def __init__(self, is_selected: bool, data_dep_view: DataDepView, node: BaseDepNode, instr: CsInsn | None) -> None:
        super().__init__()
        self.setFlags(QtWidgets.QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)

        self._selected = is_selected
        self._data_dep_view = data_dep_view
        self._workspace = self._data_dep_view.workspace
        self._node: BaseDepNode = node
        self._instr = instr

        self._header_text: str | None = None
        self._instruction_text: str | None = None
        self._y_off = 0  # Used to track height
        self._header_text_item: QtWidgets.QGraphicsSimpleTextItem | None = None
        self._instruction_text_item: QtWidgets.QGraphicsSimpleTextItem | None = None

        self._init_widgets()
        self._update_size()

        self.setAcceptHoverEvents(True)

    @property
    def selected(self) -> bool:
        return self._selected

    @selected.setter
    def selected(self, new_val: bool) -> None:
        self._selected = new_val
        self.update()

    @property
    def node(self) -> BaseDepNode:
        return self._node

    def _build_simple_text_item(self, text: str) -> QtWidgets.QGraphicsSimpleTextItem:
        text_item = QtWidgets.QGraphicsSimpleTextItem(text, self)
        text_item.setFont(Conf.symexec_font)
        text_item.setBrush(QtCore.Qt.GlobalColor.black)
        text_item.setPos(self.HORIZONTAL_PADDING, self.VERTICAL_PADDING + self._y_off)
        return text_item

    def _init_widgets(self) -> None:
        self._y_off = 0

        if not isinstance(self._node, ConstantDepNode):
            # ConstantDepNodes already display their value, by default
            self.setToolTip(f"Value: {hex(self._node.value)}")

        # Display the instruction, if applicable
        self._header_text = str(self._node)
        self._header_text_item = self._build_simple_text_item(self._header_text)
        self._y_off += self._header_text_item.boundingRect().height() + 3
        # Make header bold
        header_font = self._header_text_item.font()

        header_font.setBold(True)
        self._header_text_item.setFont(header_font)

        self._instruction_text = (
            f"{hex(self._instr.address)}: {self._instr.mnemonic} {self._instr.op_str}"
            if self._instr
            else f"{hex(self._node.ins_addr)}:{self._node.stmt_idx}"
        )  # TODO: Reset else to ''
        self._instruction_text_item = self._build_simple_text_item(self._instruction_text)
        if self._instruction_text:
            self._y_off += self._instruction_text_item.boundingRect().height() + 3

    def refresh(self) -> None:
        self._update_size()

    def _paint_boundary(self, painter: QtGui.QPainter) -> None:
        painter.setFont(Conf.symexec_font)

        # Pick background color based on node type
        if isinstance(self._node, ConstantDepNode):
            color = QDataDepGraphBlock.CONSTANT_BACKGROUND
        elif isinstance(self._node, TmpDepNode):
            color = QDataDepGraphBlock.TMP_BACKGROUND
        else:
            color = QDataDepGraphBlock.DEFAULT_BACKGROUND

        # If the node is selected, darken this color
        if self._selected:
            color = color.darker(150)
        painter.setBrush(color)

        border_color = QtGui.QColor(0, 254, 254) if self._selected else QtGui.QColor(240, 240, 240)
        painter.setPen(QtGui.QPen(border_color, 1.5))
        painter.drawRect(0, 0, self.width, self.height)

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionGraphicsItem,  # pylint: disable=unused-argument
        widget: QtWidgets.QWidget | None = ...,
    ) -> None:  # pylint: disable=unused-argument
        self._paint_boundary(painter)

    def _boundingRect(self):
        return QtCore.QRectF(0, 0, self._width, self._height)

    def _update_size(self) -> None:
        max_text_item_width = max(
            self._header_text_item.boundingRect().width(),
            self._instruction_text_item.boundingRect().width(),
        )

        self._width = self.HORIZONTAL_PADDING * 2 + max_text_item_width
        self._height = self.VERTICAL_PADDING * 2 + self._y_off

        self._width = max(30, self._width)
        self._height = max(10, self._height)
        self.recalculate_size()

    #
    # Event Handlers
    #

    def mouseDoubleClickEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        if event.button() == QtCore.Qt.MouseButton.LeftButton and self._node.ins_addr:
            self._workspace.viz(self._node.ins_addr)

    def hoverEnterEvent(self, event: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        self._selected = True
        self.refresh()
        self.setFocus(QtCore.Qt.FocusReason.MouseFocusReason)
        self.grabKeyboard()
        self._data_dep_view.hover_enter_block(self, event.modifiers())

    def hoverLeaveEvent(self, event: QtWidgets.QGraphicsSceneHoverEvent) -> None:  # pylint: disable=unused-argument
        self._selected = False
        self.refresh()
        self.ungrabKeyboard()
        self.clearFocus()
        self._data_dep_view.hover_leave_block()

    def contextMenuEvent(self, event: QtWidgets.QGraphicsSceneContextMenuEvent) -> None:
        ctxt_menu = QtWidgets.QMenu("")
        ctxt_menu.addAction("Trace backward", lambda: self._data_dep_view.use_subgraph(self, True))
        ctxt_menu.addAction("Trace forward", lambda: self._data_dep_view.use_subgraph(self, False))
        ctxt_menu.exec_(event.screenPos())

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        """Handle change from ancestor to descendant tracing"""
        if self._selected and event.key() == QtCore.Qt.Key.Key_Control:
            self._data_dep_view.update_descendants(self)
        else:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QtGui.QKeyEvent) -> None:
        """Handle change form descendant to ancestor tracing"""
        if self._selected and event.key() == QtCore.Qt.Key.Key_Control:
            self._data_dep_view.update_ancestors(self)
        else:
            super().keyReleaseEvent(event)

    def mousePressEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:  # pylint:disable=no-self-use
        event.accept()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from angr.analyses.disassembly import ConstantOperand, MemoryOperand, RegisterOperand, Value
from PySide6.QtCore import QPointF, QRectF, Qt
from PySide6.QtWidgets import QApplication, QGraphicsSimpleTextItem

from angrmanagement.logic.disassembly.info_dock import OperandDescriptor, OperandHighlightMode

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance

log = logging.getLogger(__name__)


class QOperand(QCachedGraphicsItem):
    BRANCH_TARGETS_SPACING = 5
    LABEL_VARIABLE_SPACING = 5
    VARIABLE_IDENT_SPACING = 5

    def __init__(
        self,
        instance: Instance,
        func_addr,
        disasm_view,
        disasm,
        infodock,
        insn,
        operand,
        operand_index,
        is_branch_target: bool,
        is_indirect_branch: bool,
        branch_targets,
        config,
        parent=None,
    ) -> None:
        super().__init__(parent=parent)

        self.instance = instance
        self.func_addr = func_addr
        self.disasm_view = disasm_view
        self.disasm = disasm
        self.infodock = infodock
        self.variable_manager = infodock.variable_manager
        self.insn = insn
        self.operand = operand
        self.operand_index = operand_index
        self.is_branch_target = is_branch_target
        self.is_indirect_branch = is_indirect_branch
        self.branch_targets = branch_targets

        self._branch_target: int | None = None

        # the variable involved
        self.variable = None

        self._cachy = None

        self._config = config

        # "widgets"
        self._label = None
        self._label_item: QGraphicsSimpleTextItem | None = None
        self._variable_label = None
        self._variable_label_item: QGraphicsSimpleTextItem | None = None
        self._variable_ident = None
        self._variable_ident_item: QGraphicsSimpleTextItem | None = None
        self._branch_targets = None
        self._branch_targets_text = None
        self._branch_targets_item: QGraphicsSimpleTextItem | None = None
        self._is_target_func = None

        self._width = None
        self._height = None

        self._init_widgets()

    #
    # Properties
    #

    @property
    def text(self):
        return self._label

    @property
    def is_constant(self):
        return isinstance(self.operand, ConstantOperand)

    @property
    def constant_value(self):
        if self.is_constant:
            return self.operand.cs_operand.imm
        return None

    @property
    def is_constant_memory(self):
        return (
            isinstance(self.operand, MemoryOperand)
            and len(self.operand.values) == 1
            and isinstance(self.operand.values[0], Value)
            and isinstance(self.operand.values[0].val, int)
        )

    @property
    def constant_memory_value(self):
        if self.is_constant_memory:
            return self.operand.values[0].val
        return None

    @property
    def selected(self):
        return self.infodock.is_operand_selected(self.insn.addr, self.operand_index)

    @property
    def operand_descriptor(self):
        return OperandDescriptor(
            self.text,
            self.constant_value,
            func_addr=self.func_addr,
            variable_ident=self.variable.ident if self.variable is not None else None,
        )

    #
    # Event handlers
    #

    def mousePressEvent(self, event) -> None:
        if event.button() == Qt.MouseButton.LeftButton:
            selected = self.infodock.toggle_operand_selection(
                self.insn.addr,
                self.operand_index,
                self.operand_descriptor,
                insn_pos=self.parentItem().scenePos(),
                unique=QApplication.keyboardModifiers() != Qt.KeyboardModifier.ControlModifier,
            )
            if selected:
                # select the current instruction, too
                self.infodock.select_instruction(self.insn.addr, insn_pos=QPointF(self.x(), self.y()), unique=True)
        else:
            super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event) -> None:
        button = event.button()
        if button == Qt.MouseButton.LeftButton:
            if self._branch_target is not None:
                self.disasm_view.jump_to(self._branch_target, src_ins_addr=self.insn.addr, use_animation=True)
                return
            if self.is_constant:
                self.disasm_view.jump_to(self.constant_value, src_ins_addr=self.insn.addr, use_animation=True)
                return
            if self.is_constant_memory:
                self.disasm_view.jump_to(self.constant_memory_value, src_ins_addr=self.insn.addr, use_animation=True)
        else:
            super().mouseDoubleClickEvent(event)

    #
    # Public methods
    #

    def refresh(self) -> None:
        self._init_widgets()
        self.recalculate_size()

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        # Background
        if self.selected:
            painter.setPen(self._config.disasm_view_operand_select_color)
            painter.setBrush(self._config.disasm_view_operand_select_color)
            painter.drawRect(0, 0, self.width, self.height)
        else:
            for _, selected_operand_desc in self.infodock.selected_operands.items():
                if self._equals_for_highlighting_purposes(selected_operand_desc):
                    painter.setBrush(self._config.disasm_view_operand_highlight_color)
                    painter.setPen(self._config.disasm_view_operand_highlight_color)
                    painter.drawRect(0, 0, self.width, self.height)
                    break

    #
    # Private methods
    #

    def _branch_target_for_operand(self, operand, branch_targets):
        if not branch_targets:
            return None

        if len(branch_targets) == 1:
            return next(iter(branch_targets))

        # there are more than one targets
        # we pick the one that complies with the operand's text
        # my solution is pretty hackish...

        imm = self.constant_value
        if imm is not None and imm in branch_targets:
            # problem solved
            return imm
        else:
            # umm why?
            pass

        # try to render it
        rendered = operand.render()[0]
        for t in branch_targets:
            if rendered in (f"{t:x}", f"{t:#x}"):
                return t
            if t == rendered:
                return t

        # ouch not sure what to do
        log.warning('Cannot determine branch targets for operand "%s". Please report on GitHub.', rendered)
        # return a random one
        return next(iter(branch_targets))

    @staticmethod
    def _first_n_branch_targets(branch_targets, n):
        if not branch_targets:
            return []

        return list(branch_targets)[:n]

    def _init_widgets(self) -> None:
        if self.is_branch_target:
            # a branch instruction

            is_target_func = bool(
                self.branch_targets is not None and next(iter(self.branch_targets)) in self.disasm.kb.functions
            )

            self._label = self.operand.render()[0]
            self._is_target_func = is_target_func

            if self.is_indirect_branch:
                # indirect jump
                self._branch_targets = self.branch_targets
                first_n_targets = self._first_n_branch_targets(self._branch_targets, 3)
                if first_n_targets:
                    targets = []
                    for t in first_n_targets:
                        txt = None
                        if is_target_func:
                            # try to get a function
                            try:
                                target_func = self.disasm.kb.functions.get_by_addr(t)
                                txt = target_func.demangled_name
                            except KeyError:
                                pass
                        # is the address a label?
                        if txt is None and t in self.disasm.kb.labels:
                            txt = self.disasm.kb.labels[t]
                        if txt is None:
                            # use the hex text
                            txt = f"{t:#08x}"
                        targets.append(txt)
                    self._branch_targets_text = "[ " + ", ".join(targets) + " ]"
                else:
                    self._branch_targets_text = "[ unknown ]"

                if self._branch_targets and len(self._branch_targets) == 1:
                    self._branch_target = next(iter(self._branch_targets))

            else:
                self._branch_target = self._branch_target_for_operand(self.operand, self.branch_targets)

        else:
            # not a branch

            formatting = {}
            if isinstance(self.operand, MemoryOperand):
                variable_sort = "memory"
            elif isinstance(self.operand, RegisterOperand):
                variable_sort = "register"
            else:
                variable_sort = None

            # without displaying variable
            self._label = self.operand.render(formatting=formatting)[0]

            if variable_sort and self.variable_manager is not None:
                # try find the corresponding variable
                variable_and_offsets = self.variable_manager[self.func_addr].find_variables_by_insn(
                    self.insn.addr, variable_sort
                )
                if variable_and_offsets:
                    variable, offset = self._pick_variable(variable_and_offsets)

                    if variable is not None:
                        self.variable = variable
                        self._variable_ident = f"<{variable.ident}>"
                        if offset is None:
                            offset = 0

                        variable_str = variable.name

                        ident = (self.insn.addr, "operand", self.operand_index)
                        if "custom_values_str" not in formatting:
                            formatting["custom_values_str"] = {}
                        if variable_sort == "memory":
                            custom_value_str = f"{variable_str}" if offset == 0 else f"{variable_str}[{offset}]"
                        else:
                            custom_value_str = ""

                        ##
                        # Hacks
                        ##
                        if self.infodock.induction_variable_analysis is not None:
                            r = self.infodock.induction_variable_analysis.variables.get(variable.ident, None)
                            if r is not None and r.expr.__class__.__name__ == "InductionExpr":
                                custom_value_str = f"i*{r.expr.stride}+{r.expr.init}"
                            if (
                                r is not None
                                and r.expr.__class__.__name__ == "Add"
                                and r.expr.operands[0].__class__.__name__ == "InductionExpr"
                            ):
                                custom_value_str = (
                                    f"i*{r.expr.operands[0].stride}+"
                                    "{r.expr.operands[0].init + r.expr.operands[1].value}"
                                )

                        formatting["custom_values_str"][ident] = custom_value_str

                        if "show_prefix" not in formatting:
                            formatting["show_prefix"] = {}
                        formatting["show_prefix"][ident] = "False"

                        if "values_style" not in formatting:
                            formatting["values_style"] = {}
                        formatting["values_style"][ident] = "curly"

                    # with variable displayed
                    if variable_sort == "memory":
                        self._variable_label = self.operand.render(formatting=formatting)[0]
                    else:
                        self._variable_label = ""

        if self._branch_target or self._branch_targets:
            if self._is_target_func:
                label_color = self._config.disasm_view_target_addr_color
            else:
                label_color = self._config.disasm_view_antitarget_addr_color
        else:
            if self.is_constant:
                label_color = self._config.disasm_view_operand_constant_color
            else:
                label_color = self._config.disasm_view_operand_color

        # label
        # [rax]
        if self._label_item is not None:
            self.scene().removeItem(self._label_item)
            self._label_item = None
        self._label_item = QGraphicsSimpleTextItem(self._label, self)
        self._label_item.setFont(self._config.disasm_font)
        self._label_item.setBrush(label_color)

        # variable
        # {s_10}
        if self._variable_label_item is not None:
            self.scene().removeItem(self._variable_label_item)
            self._variable_label_item = None
        if self._variable_label:
            self._variable_label_item = QGraphicsSimpleTextItem(self._variable_label, self)
            self._variable_label_item.setFont(self._config.disasm_font)
            self._variable_label_item.setBrush(self._config.disasm_view_variable_label_color)

        # additional branch targets
        if self._branch_targets_item is not None:
            self.scene().removeItem(self._branch_targets_item)
            self._branch_targets_item = None
        if self._branch_targets_text:
            self._branch_targets_item = QGraphicsSimpleTextItem(self._branch_targets_text, self)
            self._branch_targets_item.setFont(self._config.disasm_font)
            self._branch_targets_item.setBrush(self._config.disasm_view_branch_target_text_color)

        # variable identifier
        if self._variable_ident_item is not None:
            self.scene().removeItem(self._variable_ident_item)
            self._variable_ident_item = None
        if self.variable is not None and self.disasm_view.show_variable_identifier:
            self._variable_ident_item = QGraphicsSimpleTextItem(self._variable_ident, self)
            self._variable_ident_item.setFont(self._config.disasm_font)
            self._variable_ident_item.setBrush(self._config.disasm_view_variable_ident_color)

        self._layout_items_and_update_size()

    def _layout_items_and_update_size(self) -> None:
        x, y = 0, 0

        # label
        self._label_item.setPos(x, y)
        x += self._label_item.boundingRect().width()

        # variable
        if self._variable_label_item is not None:
            if self.disasm_view.show_variable:
                x += self.LABEL_VARIABLE_SPACING
                self._variable_label_item.setPos(x, y)
                x += self._variable_label_item.boundingRect().width()
                self._variable_label_item.show()
            else:
                self._variable_label_item.hide()

        # additional branch targets
        if self._branch_targets_item is not None:
            x += self.BRANCH_TARGETS_SPACING
            self._branch_targets_item.setPos(x, y)
            x += self._branch_targets_item.boundingRect().width()

        # variable identifier
        if self._variable_ident_item is not None:
            x += self.VARIABLE_IDENT_SPACING
            self._variable_ident_item.setPos(x, y)
            x += self._variable_ident_item.boundingRect().width()

        self._width = x
        self._height = self._label_item.boundingRect().height()
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)

    def _pick_variable(self, variable_and_offsets):
        """
        Pick the corresponding variable for the current operand.

        :param list variable_and_offsets:   A list of variables and the offsets into each variable.
        :return:                            A tuple of variable and the offset.
        :rtype:                             tuple
        """

        if isinstance(self.operand, MemoryOperand):
            if len(variable_and_offsets) > 1:
                log.error("Instruction %#x has two memory operands. Please report it on GitHub.", self.insn.addr)
            return variable_and_offsets[0]

        elif isinstance(self.operand, RegisterOperand):
            # there might be multiple register-type variables for an instruction. pick the right one is... not easy

            the_reg = self.operand.register
            if the_reg is None:
                # huh, it does not have a Register child
                return None, None

            reg_name = the_reg.reg
            arch = self.instance.project.arch

            if len(variable_and_offsets) == 1:
                # only one candidate...
                var, offset = variable_and_offsets[0]
                if arch.registers[reg_name][0] == var.reg:
                    return var, offset
                return None, None

            if self.operand_index > 0:
                # this is the source operand
                # which variable is read here?
                for var, offset in variable_and_offsets:
                    if arch.registers[reg_name][0] == var.reg and self._variable_has_access(
                        var, self.insn.addr, "read"
                    ):
                        return var, offset

                log.debug(
                    "Cannot find any source variable for operand %d at instruction %#x.",
                    self.operand_index,
                    self.insn.addr,
                )
                return None, None

            # this is the destination operand
            # which variable is written here?
            for var, offset in variable_and_offsets:
                if arch.registers[reg_name][0] == var.reg and self._variable_has_access(var, self.insn.addr, "write"):
                    return var, offset

            log.debug(
                "Cannot find any destination variable for operand %d at instruction %#x.",
                self.operand_index,
                self.insn.addr,
            )
            # just return the first one
            return None, None

        else:
            # what's this type? why am I here?
            log.error("_pick_variable: Unsupported operand type %s.", self.operand.__class__)

            return None, None

    def _variable_has_access(self, variable, ins_addr, access_type):
        if variable not in self.variable_manager[self.func_addr]._variable_accesses:
            log.error("Variable %s does not have any accessing records.", variable)
            return False

        accesses = self.variable_manager[self.func_addr]._variable_accesses[variable]
        return any(access.location.ins_addr == ins_addr and access.access_type == access_type for access in accesses)

    def _equals_for_highlighting_purposes(self, other):
        """

        :param OperandDescriptor other: The other operand to compare with.
        :return:
        """

        if other is None:
            return False

        highlight_mode = self.infodock.highlight_mode

        if highlight_mode == OperandHighlightMode.SAME_TEXT or self.variable is None:
            # when there is no related variable, we highlight as long as they have the same text
            return other.text == self.text
        elif (
            highlight_mode == OperandHighlightMode.SAME_IDENT
            and self.variable is not None
            and other.variable_ident is not None
        ):
            return self.func_addr == other.func_addr and self.variable.ident == other.variable_ident

        return False



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QRectF, Qt

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from PySide6.QtGui import QPainter
    from PySide6.QtWidgets import QStyleOptionGraphicsItem, QWidget

    from angrmanagement.data.instance import Instance


class QPhiVariable(QCachedGraphicsItem):
    """QPhiVariable is a graphical representation of a phi variable in the disassembly view."""

    IDENT_LEFT_PADDING = 5

    def __init__(self, instance: Instance, disasm_view, phi_variable, config, parent=None) -> None:
        """

        :param workspace:
        :param disasm_view:
        :param PhiVariable phi_variable:
        :param config:
        """

        super().__init__(parent=parent)

        # initialization
        self.instance = instance
        self.disasm_view = disasm_view
        self.phi = phi_variable.variable  # the major variable
        self.variables = phi_variable.variables  # the sub variables
        self._config = config

        self._variable_name = None
        self._variable_name_width = None
        self._variable_ident = None
        self._variable_ident_width = None

        self._subvar_names = None
        self._subvar_name_widths = None
        self._subvar_idents = None
        self._subvar_ident_widths = None

        self._width = 0
        self._height = 0

        self._init_widgets()

    #
    # Public methods
    #

    def paint(  # pylint: disable=unused-argument
        self,
        painter: QPainter,
        option: QStyleOptionGraphicsItem,
        widget: QWidget | None = None,
    ) -> None:
        if self.disasm_view.show_variable_identifier is False:
            # Phi variables are not displayed if variable identifies are hidden
            return

        x = 0

        painter.setFont(self._config.disasm_font)

        # variable name
        painter.setPen(Qt.GlobalColor.darkGreen)
        painter.drawText(x, self._config.disasm_font_ascent, self._variable_name)
        x += self._variable_name_width

        # variable ident
        if self.disasm_view.show_variable_identifier:
            x += self.IDENT_LEFT_PADDING
            painter.setPen(Qt.GlobalColor.blue)
            painter.drawText(x, self._config.disasm_font_ascent, self._variable_ident)
            x += self._variable_ident_width

        # The equal sign
        painter.setPen(Qt.GlobalColor.black)
        painter.drawText(x, self._config.disasm_font_ascent, " = ")
        x += self._config.disasm_font_width * 3
        painter.setPen(Qt.GlobalColor.darkGreen)
        painter.drawText(x, self._config.disasm_font_ascent, "\u0278(")
        x += self._config.disasm_font_width * 2

        for i, (subvar_ident, ident_width) in enumerate(
            zip(self._subvar_idents, self._subvar_ident_widths, strict=False)
        ):
            painter.setPen(Qt.GlobalColor.darkGreen)
            painter.drawText(x, self._config.disasm_font_ascent, subvar_ident)
            x += ident_width
            if i != len(self._subvar_idents) - 1:
                painter.setPen(Qt.GlobalColor.black)
                painter.drawText(x, self._config.disasm_font_ascent, ", ")
                x += 2 + self._config.disasm_font_width

        painter.setPen(Qt.GlobalColor.darkGreen)
        painter.drawText(x, self._config.disasm_font_ascent, ")")

    def refresh(self) -> None:
        super().refresh()

        self._update_size()

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        # variable name
        self._variable_name = "{%s}" % ("Unk" if not self.phi.name else self.phi.name)
        # variable ident
        self._variable_ident = "<%s>" % ("Unk" if not self.phi.ident else self.phi.ident)

        # subvariables
        self._subvar_names = []
        self._subvar_idents = []
        for subvar in self.variables:
            name = "Unk" if not subvar.name else subvar.name
            self._subvar_names.append(name)
            ident = "<%s>" % ("Unk" if not subvar.ident else subvar.ident)
            self._subvar_idents.append(ident)

        self._update_size()

    def _update_size(self) -> None:
        if self.disasm_view.show_variable_identifier is False:
            # Phi variables are not displayed if variable identifies are hidden
            self._width = 0
            self._height = 0
            return

        self._variable_name_width = len(self._variable_name) * self._config.disasm_font_width
        self._variable_ident_width = len(self._variable_ident) * self._config.disasm_font_width

        # Update widths of sub-variables
        self._subvar_name_widths = []
        for name in self._subvar_names:
            self._subvar_name_widths.append(len(name) * self._config.disasm_font_width)
        self._subvar_ident_widths = []
        for ident in self._subvar_idents:
            self._subvar_ident_widths.append(len(ident) * self._config.disasm_font_width)

        self._width = (
            self._variable_name_width
            + self._config.disasm_font_width * 5
            + sum(self._subvar_ident_widths)  # " = ", the phi sign, and the "("
            + (len(self._subvar_ident_widths) - 1) * 2 * self._config.disasm_font_width
            + self._config.disasm_font_width * 1  # ")"
        )
        if self.disasm_view.show_variable_identifier:
            self._width += self.IDENT_LEFT_PADDING + self._variable_ident_width

        self._height = self._config.disasm_font_height

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



""" Defines miscellaneous Qt-related helper classes and functions.
"""

import inspect

from qtpy import QtCore, QtGui

from traitlets import HasTraits, TraitType

#-----------------------------------------------------------------------------
# Metaclasses
#-----------------------------------------------------------------------------

MetaHasTraits = type(HasTraits)
MetaQObject = type(QtCore.QObject)

class MetaQObjectHasTraits(MetaQObject, MetaHasTraits):
    """ A metaclass that inherits from the metaclasses of HasTraits and QObject.

    Using this metaclass allows a class to inherit from both HasTraits and
    QObject. Using SuperQObject instead of QObject is highly recommended. See
    QtKernelManager for an example.
    """
    def __new__(mcls, name, bases, classdict):
        # FIXME: this duplicates the code from MetaHasTraits.
        # I don't think a super() call will help me here.
        for k,v in iter(classdict.items()):
            if isinstance(v, TraitType):
                v.name = k
            elif inspect.isclass(v):
                if issubclass(v, TraitType):
                    vinst = v()
                    vinst.name = k
                    classdict[k] = vinst
        cls = MetaQObject.__new__(mcls, name, bases, classdict)
        return cls

    def __init__(mcls, name, bases, classdict):
        # Note: super() did not work, so we explicitly call these.
        MetaQObject.__init__(mcls, name, bases, classdict)
        MetaHasTraits.__init__(mcls, name, bases, classdict)

#-----------------------------------------------------------------------------
# Classes
#-----------------------------------------------------------------------------

def superQ(QClass):
    """ Permits the use of super() in class hierarchies that contain Qt classes.

    Unlike QObject, SuperQObject does not accept a QObject parent. If it did,
    super could not be emulated properly (all other classes in the heierarchy
    would have to accept the parent argument--they don't, of course, because
    they don't inherit QObject.)

    This class is primarily useful for attaching signals to existing non-Qt
    classes. See QtKernelManagerMixin for an example.
    """
    class SuperQClass(QClass):

        def __new__(cls, *args, **kw):
            # We initialize QClass as early as possible. Without this, Qt complains
            # if SuperQClass is not the first class in the super class list.
            inst = QClass.__new__(cls)
            QClass.__init__(inst)
            return inst

        def __init__(self, *args, **kw):
            # Emulate super by calling the next method in the MRO, if there is one.
            mro = self.__class__.mro()
            for qt_class in QClass.mro():
                mro.remove(qt_class)
            next_index = mro.index(SuperQClass) + 1
            if next_index < len(mro):
                mro[next_index].__init__(self, *args, **kw)

    return SuperQClass

SuperQObject = superQ(QtCore.QObject)

#-----------------------------------------------------------------------------
# Functions
#-----------------------------------------------------------------------------

def get_font(family, fallback=None):
    """Return a font of the requested family, using fallback as alternative.

    If a fallback is provided, it is used in case the requested family isn't
    found.  If no fallback is given, no alternative is chosen and Qt's internal
    algorithms may automatically choose a fallback font.

    Parameters
    ----------
    family : str
      A font name.
    fallback : str
      A font name.

    Returns
    -------
    font : QFont object
    """
    font = QtGui.QFont(family)
    # Check whether we got what we wanted using QFontInfo, since exactMatch()
    # is overly strict and returns false in too many cases.
    font_info = QtGui.QFontInfo(font)
    if fallback is not None and font_info.family() != family:
        font = QtGui.QFont(fallback)
    return font


# -----------------------------------------------------------------------------
# Vendored from ipython_genutils
# -----------------------------------------------------------------------------
def _chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


def _find_optimal(rlist, *, separator_size, displaywidth):
    """Calculate optimal info to columnize a list of strings"""
    for nrow in range(1, len(rlist) + 1):
        chk = list(map(max, _chunks(rlist, nrow)))
        sumlength = sum(chk)
        ncols = len(chk)
        if sumlength + separator_size * (ncols - 1) <= displaywidth:
            break

    return {
        "columns_numbers": ncols,
        "rows_numbers": nrow,
        "columns_width": chk,
    }


def _get_or_default(mylist, i, *, default):
    """return list item number, or default if don't exist"""
    if i >= len(mylist):
        return default
    else:
        return mylist[i]


def compute_item_matrix(items, empty=None, *, separator_size=2, displaywidth=80):
    """Returns a nested list, and info to columnize items

    Parameters
    ----------
    items
        list of strings to columnize
    empty : (default None)
        Default value to fill list if needed
    separator_size : int (default=2)
        How much characters will be used as a separation between each column.
    displaywidth : int (default=80)
        The width of the area onto which the columns should enter

    Returns
    -------

    strings_matrix

        nested list of strings, the outer most list contains as many list as
        rows, the innermost lists have each as many element as column. If the
        total number of elements in `items` does not equal the product of
        rows*columns, the last element of some lists are filled with `None`.

    dict_info
        Some info to make columnize easier:

        columns_numbers
          number of columns
        rows_numbers
          number of rows
        columns_width
          list of width of each columns

    Examples
    --------
    ::

        In [1]: l = ['aaa','b','cc','d','eeeee','f','g','h','i','j','k','l']
           ...: compute_item_matrix(l,displaywidth=12)
        Out[1]:
            ([['aaa', 'f', 'k'],
            ['b', 'g', 'l'],
            ['cc', 'h', None],
            ['d', 'i', None],
            ['eeeee', 'j', None]],
            {'columns_numbers': 3,
            'columns_width': [5, 1, 1],
            'rows_numbers': 5})
    """
    info = _find_optimal(
        [len(it) for it in items], separator_size=separator_size, displaywidth=displaywidth
    )
    nrow, ncol = info["rows_numbers"], info["columns_numbers"]
    return (
        [
            [_get_or_default(items, c * nrow + i, default=empty) for c in range(ncol)]
            for i in range(nrow)
        ],
        info,
    )


def columnize(items, separator="  ", displaywidth=80):
    """Transform a list of strings into a single string with columns.

    Parameters
    ----------
    items : sequence of strings
        The strings to process.

    Returns
    -------
    The formatted string.
    """
    if not items:
        return "\n"
    matrix, info = compute_item_matrix(
        items, separator_size=len(separator), displaywidth=displaywidth
    )
    fmatrix = [filter(None, x) for x in matrix]
    sjoin = lambda x: separator.join(
        [y.ljust(w, " ") for y, w in zip(x, info["columns_width"])]
    )
    return "\n".join(map(sjoin, fmatrix)) + "\n"


def import_item(name):
    """Import and return ``bar`` given the string ``foo.bar``.

    Calling ``bar = import_item("foo.bar")`` is the functional equivalent of
    executing the code ``from foo import bar``.

    Parameters
    ----------
    name : string
      The fully qualified name of the module/package being imported.

    Returns
    -------
    mod : module object
       The module that was imported.
    """
    parts = name.rsplit(".", 1)

    if len(parts) == 2:
        # called with 'foo.bar....'
        package, obj = parts
        module = __import__(package, fromlist=[obj])
        try:
            pak = getattr(module, obj)
        except AttributeError:
            raise ImportError("No module named %s" % obj)
        return pak
    else:
        # called with un-dotted string
        return __import__(parts[0])



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from angr.analyses.cfg.cfb import MemoryRegion, Unknown
from angr.block import Block
from angr.knowledge_plugins.cfg.memory_data import MemoryData
from PySide6.QtCore import QEvent, QRect, QRectF, Qt
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QAbstractScrollArea, QAbstractSlider, QGraphicsScene, QHBoxLayout
from sortedcontainers import SortedDict

from angrmanagement.config import Conf

from .qblock import QLinearBlock
from .qdisasm_base_control import DisassemblyLevel, QDisassemblyBaseControl
from .qgraph import QSaveableGraphicsView
from .qmemory_data_block import QMemoryDataBlock
from .qunknown_block import QUnknownBlock

if TYPE_CHECKING:
    from angr.analyses import Disassembly
    from angr.analyses.decompiler import Clinic
    from angr.knowledge_plugins import Function

    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.disassembly import InfoDock


_l = logging.getLogger(__name__)


class QLinearDisassemblyView(QSaveableGraphicsView):
    def __init__(self, area, parent=None) -> None:
        super().__init__(parent=parent)

        self.area: QLinearDisassembly = area
        self._scene = QGraphicsScene(0, 0, self.width(), self.height())
        self.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        self.setScene(self._scene)
        self.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)

        # Do not use the scrollbars since they are hard-linked to the size of the scene, which is bad for us
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    def wheelEvent(self, event) -> None:
        self.area.wheelEvent(event)
        super().wheelEvent(event)

    def event(self, event):
        """
        Reimplemented to capture the Tab keypress event.
        """

        # by default, the tab key moves focus. Hijack the tab key
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Tab:
            self.area.disasm_view.keyPressEvent(event)
            return True
        return super().event(event)


class QLinearDisassembly(QDisassemblyBaseControl, QAbstractScrollArea):
    OBJECT_PADDING = 0

    def __init__(self, instance: Instance, disasm_view, parent=None) -> None:
        QDisassemblyBaseControl.__init__(self, instance, disasm_view, QAbstractScrollArea)
        QAbstractScrollArea.__init__(self, parent=parent)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        self.horizontalScrollBar().setSingleStep(Conf.disasm_font_width)
        self.verticalScrollBar().setSingleStep(16)

        # self.setTransformationAnchor(QGraphicsView.NoAnchor)
        # self.setResizeAnchor(QGraphicsView.NoAnchor)
        # self.setAlignment(Qt.AlignLeft)

        self._viewer: QLinearDisassemblyView

        self._line_height = Conf.disasm_font_height

        self._offset_to_region = SortedDict()
        self._addr_to_region_offset = SortedDict()
        # Offset (in bytes) into the entire blanket view
        self._offset = 0
        # The maximum offset (in bytes) of the blanket view
        self._max_offset = None
        # The first line that is rendered of the first object in self.objects. Start from 0.
        self._start_line_in_object = 0

        self._disasms = {}
        self._ail_disasms = {}
        self.objects = []

        self.verticalScrollBar().actionTriggered.connect(self._on_vertical_scroll_bar_triggered)

        self._init_widgets()
        self.initialize()

    def reload(self, old_infodock: InfoDock | None = None) -> None:
        curr_offset = self._offset
        self.initialize()
        self._offset = None  # force a re-generation of objects
        self.prepare_objects(curr_offset, start_line=self._start_line_in_object)
        self.redraw()

    #
    # Properties
    #

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, v) -> None:
        self._offset = v

    @property
    def max_offset(self):
        if self._max_offset is None:
            self._max_offset = self._calculate_max_offset()
        return self._max_offset

    @property
    def cfg(self):
        return self.instance.cfg

    @property
    def cfb(self):
        return self.instance.cfb

    @property
    def scene(self):
        return self._viewer._scene

    #
    # Events
    #

    def resizeEvent(self, event) -> None:
        old_height = event.oldSize().height()
        new_height = event.size().height()
        self._viewer._scene.setSceneRect(QRectF(0, 0, event.size().width(), new_height))

        if new_height > old_height:
            # we probably need more objects generated
            curr_offset = self._offset
            self._offset = None  # force a re-generation of objects
            self.prepare_objects(curr_offset, start_line=self._start_line_in_object)
            self.redraw()

        super().resizeEvent(event)

    def wheelEvent(self, event) -> None:
        """
        :param QWheelEvent event:
        :return:
        """
        delta = event.angleDelta().y()
        if delta < 0:
            # scroll down by some lines
            lines = min(int(-delta // self._line_height), 3)
            self.prepare_objects(self.offset, start_line=self._start_line_in_object + lines)
        elif delta > 0:
            # Scroll up by some lines
            lines = min(int(delta // self._line_height), 3)
            self.prepare_objects(self.offset, start_line=self._start_line_in_object - lines)

        self.verticalScrollBar().setValue(self.offset * self._line_height)
        event.accept()
        self.viewport().update()

    def changeEvent(self, event: QEvent) -> None:
        """
        Redraw on color scheme update.
        """
        if event.type() == QEvent.Type.PaletteChange:
            self.reload()

    def _on_vertical_scroll_bar_triggered(self, action) -> None:
        action = QAbstractSlider.SliderAction(action)  # XXX: `action` is passed as an int

        if action == QAbstractSlider.SliderAction.SliderSingleStepAdd:
            # scroll down by one line
            self.prepare_objects(self.offset, start_line=self._start_line_in_object + 1)
            self.viewport().update()
        elif action == QAbstractSlider.SliderAction.SliderSingleStepSub:
            # Scroll up by one line
            self.prepare_objects(self.offset, start_line=self._start_line_in_object - 1)
            self.viewport().update()
        elif action == QAbstractSlider.SliderAction.SliderPageStepAdd:
            # Scroll down by one page
            lines_per_page = int(self.height() // self._line_height)
            self.prepare_objects(self.offset, start_line=self._start_line_in_object + lines_per_page)
            self.viewport().update()
        elif action == QAbstractSlider.SliderAction.SliderPageStepSub:
            # Scroll up by one page
            lines_per_page = int(self.height() // self._line_height)
            self.prepare_objects(self.offset, start_line=self._start_line_in_object - lines_per_page)
            self.viewport().update()
        elif action == QAbstractSlider.SliderAction.SliderMove:
            # Setting a new offset
            new_offset = int(self.verticalScrollBar().value() // self._line_height)
            self.prepare_objects(new_offset)
            self.viewport().update()

    #
    # Public methods
    #

    def redraw(self) -> None:
        if self._viewer is not None:
            self._viewer.redraw()

    def refresh(self) -> None:
        self._update_size()
        self.redraw()

    def initialize(self) -> None:
        if self.cfb.am_none:
            return

        self._addr_to_region_offset.clear()
        self._offset_to_region.clear()
        self._disasms.clear()
        self._ail_disasms.clear()
        self._offset = None
        self._max_offset = None
        self._start_line_in_object = 0

        # enumerate memory regions
        byte_offset = 0
        mr: MemoryRegion
        for mr in self.cfb.regions:
            if mr.type in {"tls", "kernel"}:
                # Skip TLS objects and kernel objects
                continue
            self._addr_to_region_offset[mr.addr] = byte_offset
            self._offset_to_region[byte_offset] = mr
            byte_offset += mr.size

        self.refresh()

    def goto_function(self, func) -> None:
        if func.addr not in self._block_addr_map:
            _l.error("Unable to find entry block for function %s", func)
        view_height = self.viewport().height()
        desired_center_y = self._block_addr_map[func.addr].pos().y()
        _l.debug("Going to function at 0x%x by scrolling to %s", func.addr, desired_center_y)
        self.verticalScrollBar().setValue(desired_center_y - (view_height / 3))

    def show_instruction(
        self,
        insn_addr,
        insn_pos=None,
        centering: bool = False,
        use_block_pos: bool = False,
        use_animation: bool = False,
    ) -> None:
        """

        :param insn_addr:
        :param QGraphicsItem item:
        :param centering:
        :param use_block_pos:
        :return:
        """
        if insn_pos is not None:
            # check if item is already visible in the viewport
            viewport = self._viewer.viewport()
            rect = self._viewer.mapToScene(QRect(0, 0, viewport.width(), viewport.height())).boundingRect()
            if rect.contains(insn_pos):
                return

        self.navigate_to_addr(insn_addr)

    def navigate_to_addr(self, addr: int) -> None:
        if not self._addr_to_region_offset:
            return
        try:
            floor_region_addr = next(self._addr_to_region_offset.irange(maximum=addr, reverse=True))
        except StopIteration:
            floor_region_addr = next(self._addr_to_region_offset.irange())
        floor_region_offset = self._addr_to_region_offset[floor_region_addr]

        offset_into_region = addr - floor_region_addr
        self.navigate_to(floor_region_offset + offset_into_region)

    def navigate_to(self, offset: int) -> None:
        self.verticalScrollBar().setValue(offset * self._line_height)
        self.prepare_objects(offset, start_line=0)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self._viewer = QLinearDisassemblyView(self)

        layout = QHBoxLayout()
        layout.addWidget(self._viewer)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def _update_size(self) -> None:
        # ask all objects to update their sizes
        for obj in self.objects:
            obj.clear_cache()
            obj.refresh()

        # update vertical scrollbar
        self.verticalScrollBar().setRange(0, self.max_offset * self._line_height - self.height() // 2)
        offset = 0 if self.offset is None else self.offset
        self.verticalScrollBar().setValue(offset * self._line_height)

    def clear_objects(self) -> None:
        self.objects.clear()
        self._offset = None

    def prepare_objects(self, offset: int, start_line: int = 0) -> None:
        """
        Prepare objects to print based on offset and start_line. Update self.objects, self._offset, and
        self._start_line_in_object.
        :param int offset:      Beginning offset (in bytes) to display in the linear viewer.
        :param int start_line:  The first line into the first object to display in the linear viewer.
        :return:                None
        """

        if offset is None:
            offset = 0

        if offset == self._offset and start_line == self._start_line_in_object:
            return

        # make sure self._offset is not None after return
        self._offset = offset

        # Convert the offset to memory region
        base_offset: int
        mr: MemoryRegion
        base_offset, mr = self._region_from_offset(offset)
        if mr is None:
            return

        addr = self._addr_from_offset(mr, base_offset, offset)
        _l.debug("Address %#x, offset %d, start_line %d.", addr, offset, start_line)

        self._insaddr_to_block.clear()
        if start_line < 0:
            # Which object are we currently displaying at the top of the disassembly view?
            try:
                top_obj_addr = self.cfb.floor_addr(addr=addr)
            except KeyError:
                top_obj_addr = addr

            # Reverse-iterate until we have enough lines to compensate start_line
            for obj_addr, obj in self.cfb.ceiling_items(addr=top_obj_addr, reverse=True, include_first=False):
                qobject = self._obj_to_paintable(obj_addr, obj)
                if qobject is None:
                    continue
                object_lines = int(qobject.height // self._line_height)
                _l.debug("Compensating negative start_line: object %s, object_lines %d.", obj, object_lines)
                start_line += object_lines
                if start_line >= 0:
                    addr = obj_addr
                    # Update offset
                    new_region_addr = next(self._addr_to_region_offset.irange(maximum=addr, reverse=True))
                    new_region_offset = self._addr_to_region_offset[new_region_addr]
                    offset = (addr - new_region_addr) + new_region_offset
                    break
            else:
                # umm we don't have enough objects to compensate the negative start_line
                start_line = 0
                # update addr and offset to their minimal values
                addr = next(self._addr_to_region_offset.irange())
                offset = self._addr_to_region_offset[addr]

        _l.debug("After adjustment: Address %#x, offset %d, start_line %d.", addr, offset, start_line)

        scene = self.scene
        # remove existing objects
        for obj in self.objects:
            scene.removeItem(obj)
        self.objects = []

        viewable_lines = int(self.height() // self._line_height)
        lines = 0
        start_line_in_object = 0

        # Load a page of objects
        x = 80
        y = -start_line * self._line_height

        for obj_addr, obj in self.cfb.floor_items(addr=addr):
            qobject = self._obj_to_paintable(obj_addr, obj)
            _l.debug("Converted %s to %s at %x.", obj, qobject, obj_addr)
            if qobject is None:
                # Conversion failed
                continue

            if isinstance(qobject, QLinearBlock):
                for insn_addr in qobject.addr_to_insns:
                    self._insaddr_to_block[insn_addr] = qobject

            # qobject.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

            object_lines = int(qobject.height // self._line_height)

            if start_line > 0 and start_line >= object_lines:
                # this object should be skipped. ignore it
                start_line -= object_lines
                # adjust the offset as well
                if obj_addr <= addr < obj_addr + obj.size:
                    offset += obj_addr + obj.size - addr
                else:
                    offset += obj.size
                _l.debug("Skipping object %s (size %d). New offset: %d.", obj, obj.size, offset)
                y = -start_line * self._line_height
            else:
                if start_line > 0:
                    _l.debug(
                        "First object to paint: %s (size %d). Current offset %d. Start printing from line %d. "
                        "Y pos %d.",
                        obj,
                        obj.size,
                        offset,
                        start_line,
                        y,
                    )
                    # this is the first object to paint
                    start_line_in_object = start_line
                    start_line = 0
                    lines += object_lines - start_line_in_object
                else:
                    lines += object_lines
                self.objects.append(qobject)
                qobject.setPos(x, y)
                _l.debug("Adding object %s (height %s) at position %d, %d.", qobject, qobject.height, x, y)
                scene.addItem(qobject)
                y += qobject.height + self.OBJECT_PADDING

            if lines > viewable_lines:
                break

        _l.debug("Final offset %d, start_line_in_object %d.", offset, start_line_in_object)

        # Update properties
        self._offset = offset
        self._start_line_in_object = start_line_in_object

    def _obj_to_paintable(self, obj_addr, obj):
        if isinstance(obj, Block):
            cfg_node = self.cfg.get_any_node(obj_addr, force_fastpath=True)
            if cfg_node is not None:
                func_addr = cfg_node.function_address
                if self.instance.kb.functions.contains_addr(func_addr):
                    func = self.instance.kb.functions[func_addr]
                    disasm = self._get_disasm(func)
                    qobject = None
                    if self._disassembly_level is DisassemblyLevel.AIL:
                        ail_obj = None
                        if disasm.graph is not None:
                            # Clinic may yield a None graph when the function is empty
                            for n in disasm.graph.nodes:
                                if n.addr == obj.addr:
                                    ail_obj = n
                            # the corresponding AIL block may not exist
                            if ail_obj is not None:
                                qobject = QLinearBlock(
                                    self.instance,
                                    func_addr,
                                    self.disasm_view,
                                    disasm,
                                    self.disasm_view.infodock,
                                    obj.addr,
                                    ail_obj,
                                    None,
                                    None,
                                )
                    else:
                        qobject = QLinearBlock(
                            self.instance,
                            func_addr,
                            self.disasm_view,
                            disasm,
                            self.disasm_view.infodock,
                            obj.addr,
                            [obj],
                            {},
                            None,
                        )
                else:
                    # TODO: Get disassembly even if the function does not exist
                    _l.warning(
                        "Function %s does not exist, and we cannot get disassembly for block %s.", func_addr, obj
                    )
                    qobject = None
            else:
                _l.warning("Failed to get a CFG node for address %#x.", obj_addr)
                qobject = None
        elif isinstance(obj, MemoryData):
            qobject = QMemoryDataBlock(self.instance, self.disasm_view.infodock, obj_addr, obj, parent=None)
        elif isinstance(obj, Unknown):
            qobject = QUnknownBlock(self.instance, obj_addr, obj.bytes)
        else:
            qobject = None
        return qobject

    def _calculate_max_offset(self):
        try:
            max_off = next(self._offset_to_region.irange(reverse=True))
            mr: MemoryRegion = self._offset_to_region[max_off]
            return max_off + mr.size
        except StopIteration:
            return 0

    def _region_from_offset(self, offset: int):
        try:
            off = next(self._offset_to_region.irange(maximum=offset, reverse=True))
            return off, self._offset_to_region[off]
        except StopIteration:
            return None, None

    def _addr_from_offset(self, mr, base_offset, offset: int):
        return mr.addr + (offset - base_offset)

    def _get_disasm(self, func: Function) -> Clinic | Disassembly | None:
        """
        Get disassembly analysis object for a given function
        """
        if self._disassembly_level is DisassemblyLevel.AIL:
            if func.addr not in self._ail_disasms:
                self._ail_disasms[func.addr] = self.instance.project.analyses.Clinic(func)
            return self._ail_disasms[func.addr]

        if func.addr not in self._disasms:
            include_ir = self._disassembly_level is DisassemblyLevel.LifterIR
            self._disasms[func.addr] = self.instance.project.analyses.Disassembly(function=func, include_ir=include_ir)
        return self._disasms[func.addr]



from __future__ import annotations

from .qaddress_input import QAddressInput
from .qdisasm_base_control import DisassemblyLevel

# graphs
from .qdisasm_graph import QDisassemblyGraph
from .qdisasm_statusbar import QDisasmStatusBar

# other widgets
from .qfeature_map import QFeatureMap
from .qicon_label import QIconLabel
from .qinst_annotation import QAvoidAddrAnnotation, QBlockAnnotations, QFindAddrAnnotation, QHookAnnotation
from .qlinear_viewer import QLinearDisassembly, QLinearDisassemblyView
from .qstate_combobox import QStateComboBox
from .qsymexec_graph import QSymExecGraph
from .qtrace_map import QTraceMap

__all__ = [
    "DisassemblyLevel",
    "QAddressInput",
    "QAvoidAddrAnnotation",
    "QBlockAnnotations",
    "QDisasmStatusBar",
    "QDisassemblyGraph",
    "QFeatureMap",
    "QFindAddrAnnotation",
    "QHookAnnotation",
    "QIconLabel",
    "QLinearDisassembly",
    "QLinearDisassemblyView",
    "QStateComboBox",
    "QSymExecGraph",
    "QTraceMap",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from angrmanagement.logic import GlobalInfo

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class DecompileFunctionJob(InstanceJob):
    """
    The job for running the decompiler analysis. You can trigger this by pressing f5 in a function.
    """

    def __init__(self, instance: Instance, function, on_finish=None, blocking: bool = False, **kwargs) -> None:
        super().__init__("Decompiling", instance, on_finish=on_finish, blocking=blocking)
        self.kwargs = kwargs
        self.function = function

    def run(self, ctx: JobContext) -> None:
        decompiler = self.instance.project.analyses.Decompiler(
            self.function,
            flavor="pseudocode",
            variable_kb=self.instance.pseudocode_variable_kb,
            **self.kwargs,
            progress_callback=ctx.set_progress,
        )
        # cache the result
        self.instance.kb.decompilations[(self.function.addr, "pseudocode")] = decompiler.cache

        GlobalInfo.main_window.workspace.plugins.decompile_callback(self.function)



# pylint:disable=unused-argument
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QAbstractTableModel, Qt
from PySide6.QtGui import QClipboard, QCursor, QGuiApplication, QIcon, QKeySequence
from PySide6.QtWidgets import QAbstractItemView, QTableView

from angrmanagement.consts import IMG_LOCATION
from angrmanagement.logic.threads import gui_thread_schedule_async
from angrmanagement.ui.menus.log_menu import LogMenu

if TYPE_CHECKING:
    import PySide6

    from angrmanagement.data.log import LogRecord


class QLogIcons:
    """
    All icons that are used in the log table.
    """

    BENCHMARK = None
    WARNING = None
    ERROR = None

    @classmethod
    def benchmark(cls) -> QIcon:
        if cls.BENCHMARK is None:
            cls.BENCHMARK = QIcon(os.path.join(IMG_LOCATION, "benchmark-icon.png"))
        return cls.BENCHMARK

    @classmethod
    def warning(cls) -> QIcon:
        if cls.WARNING is None:
            cls.WARNING = QIcon(os.path.join(IMG_LOCATION, "warning-icon.png"))
        return cls.WARNING

    @classmethod
    def error(cls) -> QIcon:
        if cls.ERROR is None:
            cls.ERROR = QIcon(os.path.join(IMG_LOCATION, "error-icon.png"))
        return cls.ERROR


class QLogTableModel(QAbstractTableModel):
    """
    Implements a table model for log items.
    """

    Headers = ["", "Timestamp", "Source", "Content"]
    COL_ICON = 0
    COL_TIMESTAMP = 1
    COL_SOURCE = 2
    COL_CONTENT = 3

    def __init__(self, log_widget: QLogWidget = None) -> None:
        super().__init__()
        self._log_widget = log_widget
        self._log: list[LogRecord] = []

    @property
    def log(self) -> list[LogRecord]:
        return self._log

    def rowCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:
        return len(self._log)

    def columnCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:
        return len(self.Headers)

    def headerData(self, section: int, orientation: PySide6.QtCore.Qt.Orientation, role: int = ...) -> Any:
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: PySide6.QtCore.QModelIndex, role: int = ...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.log):
            return None
        log = self.log[row]
        col = index.column()

        if role == Qt.ItemDataRole.DisplayRole:
            return self._get_column_text(log, col)
        elif role == Qt.ItemDataRole.DecorationRole and col == QLogTableModel.COL_ICON:
            return self._get_column_icon(log)
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft

        return None

    @staticmethod
    def _get_column_text(log: LogRecord, col: int) -> Any:
        mapping = {
            QLogTableModel.COL_TIMESTAMP: lambda x: str(x.timestamp),
            QLogTableModel.COL_SOURCE: lambda x: str(x.source),
            QLogTableModel.COL_CONTENT: lambda x: str(x.content),
        }
        func = mapping.get(col)
        if func is None:
            return None
        return func(log)

    @staticmethod
    def _get_column_icon(log: LogRecord) -> QIcon | None:
        mapping = {
            1: QLogIcons.benchmark(),
            logging.WARNING: QLogIcons.warning(),
            logging.ERROR: QLogIcons.error(),
            logging.CRITICAL: QLogIcons.error(),
        }
        return mapping.get(log.level, None)

    @staticmethod
    def level_to_text(loglevel: int) -> str:
        mapping = {
            1: "BENCHMARK",
            logging.DEBUG: "DEBUG",
            logging.INFO: "INFO",
            logging.WARNING: "WARNING",
            logging.ERROR: "ERROR",
            logging.CRITICAL: "CRITICAL",
        }
        return mapping.get(loglevel, "")


class QLogWidget(QTableView):
    """
    Log table. Displays log messages.
    """

    def __init__(self, log_view) -> None:
        super().__init__()

        self.log_view = log_view
        # True if we want the log panel to automatically scroll to bottom
        self._auto_scroll: bool = True
        self._context_menu = LogMenu(self).qmenu()

        hheader = self.horizontalHeader()
        vheader = self.verticalHeader()
        hheader.setVisible(True)
        hheader.setStretchLastSection(True)
        vheader.setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)
        vheader.setDefaultSectionSize(20)
        self.setShowGrid(False)

        self.model: QLogTableModel = QLogTableModel(self)

        self.setModel(self.model)

        self.setColumnWidth(0, 20)
        self.setColumnWidth(1, 150)

        self.doubleClicked.connect(self._on_double_clicked)

        self.log_view.instance.log.am_subscribe(self._on_new_logrecord)

    #
    # Public methods
    #

    def clear_log(self) -> None:
        self.log_view.instance.log.am_obj = []
        self.log_view.instance.log.am_event()

    def copy_selected(self) -> None:
        content = []
        selection = self.selectionModel().selectedRows()
        for row_index in selection:
            record = self.model.log[row_index.row()]
            content.append(
                f"{QLogTableModel.level_to_text(record.level)} | "
                f"{str(record.timestamp)} | "
                f"{record.source} | "
                f"{record.content}"
            )
        self._copy_to_clipboard(os.linesep.join(content))

    def copy_selected_messages(self) -> None:
        content = []
        selection = self.selectionModel().selectedRows()
        for row_index in selection:
            content.append(self.model.log[row_index.row()].content)
        self._copy_to_clipboard(os.linesep.join(content))

    def copy_all(self) -> None:
        content = []
        for record in self.model.log:
            content.append(
                f"{QLogTableModel.level_to_text(record.level)} | "
                f"{str(record.timestamp)} | "
                f"{record.source} | "
                f"{record.content}"
            )
        self._copy_to_clipboard(os.linesep.join(content))

    def copy_all_messages(self) -> None:
        content = [record.content for record in self.model.log]
        self._copy_to_clipboard(os.linesep.join(content))

    #
    # Events
    #

    def closeEvent(self, event) -> None:
        self.log_view.instance.log.am_unsubscribe(self._on_new_logrecord)
        super().closeEvent(event)

    def contextMenuEvent(self, arg__1: PySide6.QtGui.QContextMenuEvent) -> None:
        self._context_menu.popup(QCursor.pos())

    def _on_new_logrecord(self, log_record: LogRecord = None) -> None:
        gui_thread_schedule_async(self._on_new_logrecord_core, (log_record,))

    def _on_new_logrecord_core(self, log_record: LogRecord = None) -> None:
        self._before_row_insert()

        if log_record is None:
            # reload
            self.model.layoutAboutToBeChanged.emit()
            self.model._log = self.log_view.instance.log[::]
            self.model.layoutChanged.emit()
        else:
            log_records = len(self.model.log)
            self.model.rowsAboutToBeInserted.emit(self, log_records, log_records)
            self.model.log.append(log_record)
            self.model.rowsInserted.emit(self, log_records, log_records)

        self._after_row_insert()

    def _before_row_insert(self) -> None:
        scrollbar = self.verticalScrollBar()
        self._auto_scroll = scrollbar.value() == scrollbar.maximum()

    def _after_row_insert(self) -> None:
        if self._auto_scroll:
            self.scrollToBottom()

    def keyPressEvent(self, event: PySide6.QtGui.QKeyEvent) -> None:
        if event.matches(QKeySequence.StandardKey.Copy):
            self.copy_selected_messages()
        else:
            super().keyPressEvent(event)

    def _on_double_clicked(self, item) -> None:
        # Expand/collapse row
        if self.rowHeight(item.row()) > 20:
            self.setRowHeight(item.row(), 20)
        else:
            self.setRowHeight(item.row(), self.sizeHintForRow(item.row()))

    #
    # Private methods
    #

    @staticmethod
    def _copy_to_clipboard(content: str) -> None:
        clipboard = QGuiApplication.clipboard()
        clipboard.setText(content, QClipboard.Mode.Clipboard)
        if clipboard.supportsSelection():
            clipboard.setText(content, QClipboard.Mode.Selection)



from __future__ import annotations

from angrmanagement.data.jobs import DecompileFunctionJob, VariableRecoveryJob
from angrmanagement.ui.views.code_view import CodeView
from angrmanagement.ui.views.disassembly_view import DisassemblyView


class DiffDisassemblyView(DisassemblyView):
    """
    A Disassembly View for a binary being Diffed. Should never try to synchronize normally since
    it will almost certainly have different addresses
    """

    def on_synchronized_cursor_address_changed(self) -> None:
        assert not self._processing_synchronized_cursor_update
        self._processing_synchronized_cursor_update = True
        try:
            if self.sync_state.cursor_address is not None:
                self.instance.recompilation_plugin.syncronize_with_original_disassembly_view()
        finally:
            self._processing_synchronized_cursor_update = False

    def set_synchronized_cursor_address(self, address: int | None) -> None:
        pass

    def decompile_current_function(self) -> None:
        if self.function.am_obj is not None:
            try:
                curr_ins = next(iter(self.infodock.selected_insns))
            except StopIteration:
                curr_ins = None

            view = self.workspace._get_or_create_view("pseudocode_diff", DiffCodeView)

            view.function.am_obj = self.function.am_obj
            view.function.am_event(focus=True, focus_addr=curr_ins)


class DiffCodeView(CodeView):
    """
    A view for Psuedocode of a function being Diffed.
    """

    def __init__(self, workspace, category, diff_instance, after_ready):
        super().__init__(workspace, category, diff_instance)
        self.after_ready = after_ready

    def _on_codegen_changes(self, already_regenerated: bool = False, event: str | None = None, **kwargs) -> None:
        super()._on_codegen_changes(already_regenerated, event, **kwargs)
        self.after_ready()

    def _on_new_function(self, *args, **kwargs) -> None:
        super()._on_new_function(*args, **kwargs)
        self.after_ready()

    def decompile(
        self,
        clear_prototype: bool = True,
        focus: bool = False,
        focus_addr=None,
        flavor: str = "pseudocode",
        reset_cache: bool = False,
        regen_clinic: bool = True,
    ) -> None:
        if self._function.am_none:
            return

        if clear_prototype:
            # clear the existing function prototype
            self._function.prototype = None
            self._function.ran_cca = False

        if reset_cache:
            self.instance.kb.decompilations.discard((self._function.addr, flavor))
            variables = self.instance.pseudocode_variable_kb.variables
            if variables.has_function_manager(self._function.addr):
                del variables[self._function.addr]

        def decomp_ready(*args, **kwargs) -> None:  # pylint:disable=unused-argument
            # this code is _partially_ duplicated from _on_new_function. be careful!
            available = self.instance.kb.decompilations.available_flavors(self._function.addr)
            self._update_available_views(available)
            if available:
                chosen_flavor = flavor if flavor in available else available[0]
                self.codegen.am_obj = self.instance.kb.decompilations[(self._function.addr, chosen_flavor)].codegen
                self.codegen.am_event(already_regenerated=True)
                self._focus_core(focus, focus_addr)
                if focus_addr is not None:
                    self.jump_history.record_address(focus_addr)
                else:
                    self.jump_history.record_address(self._function.am_obj.addr)
                self.after_ready()

        def decomp(*_) -> None:
            job = DecompileFunctionJob(
                self.instance,
                self._function.am_obj,
                cfg=self.instance.cfg,
                options=self._options.option_and_values,
                optimization_passes=self._options.selected_passes,
                peephole_optimizations=self._options.selected_peephole_opts,
                vars_must_struct=self.vars_must_struct,
                on_finish=decomp_ready,
                blocking=True,
                regen_clinic=regen_clinic,
            )
            self.workspace.job_manager.add_job(job)

        if self._function.ran_cca is False:
            # run calling convention analysis for this function
            if self.instance._analysis_configuration:
                options = self.instance._analysis_configuration["varec"].to_dict()
            else:
                options = {}
            options["workers"] = 0
            varrec_job = VariableRecoveryJob(self.instance, **options, on_finish=decomp, func_addr=self._function.addr)
            self.workspace.job_manager.add_job(varrec_job)
        else:
            decomp()



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QEvent, QSize, Signal
from PySide6.QtWidgets import QHBoxLayout, QLabel, QWidget

if TYPE_CHECKING:
    from PySide6.QtGui import QIcon


class QIconLabel(QWidget):
    """
    Show a label with an icon on the left.
    """

    clicked = Signal()

    def __init__(self, icon: QIcon, text: str = "") -> None:
        super().__init__()
        lyt = QHBoxLayout()
        lyt.setContentsMargins(0, 0, 0, 0)

        self._icon = icon
        self._icon_label = QLabel()
        self._update_icon()
        lyt.addWidget(self._icon_label)

        self._text_label = QLabel(text)
        lyt.addWidget(self._text_label)

        self.setLayout(lyt)
        self._update_visibility()

    def changeEvent(self, event):
        if event.type() == QEvent.Type.PaletteChange:
            self._update_icon()

    def mouseReleaseEvent(self, _) -> None:
        self.clicked.emit()

    def setText(self, text: str) -> None:
        self._text_label.setText(text)
        self._update_visibility()

    def _update_visibility(self) -> None:
        self._text_label.setVisible(self._text_label.text() != "")

    def _update_icon(self):
        self._icon_label.setPixmap(self._icon.pixmap(QSize(16, 16)))



from __future__ import annotations

import os
import sys


def is_pyinstaller() -> bool:
    """
    Detect if we are currently running as a PyInstaller-packaged program.
    :return:    True if we are running as a PyInstaller-packaged program. False if we are running in Python directly
                (e.g., development mode).
    """
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def app_path(pythonw=None, as_list: bool = False) -> str | list[str]:
    """
    Return the path of the application.

    - In standalone mode (a PyInstaller module), we return the absolute path to the executable.
    - In development mode, we return the absolute path to the python executable and "-m angr management"

    :return:    A string that represents the path to the application that can be used to run angr management.
    :rtype:     str|list
    """

    if is_pyinstaller():
        # running as a PyInstaller bundle
        if as_list:
            return [sys.executable]
        return sys.executable
    else:
        # running as a Python package
        python_path = os.path.normpath(sys.executable)
        if sys.platform.startswith("win"):
            # if pythonw is None, we don't do anything
            if pythonw is True:
                python_path = python_path.replace("python.exe", "pythonw.exe")
            elif pythonw is False:
                python_path = python_path.replace("pythonw.exe", "python.exe")
        if as_list:
            return [python_path, "-m", "angrmanagement"]
        else:
            if " " in python_path:
                python_path = f'"{python_path}"'
            app_path = python_path + " -m angrmanagement"
            return app_path


def app_root() -> str:
    """
    Return the path of the application.

    - In standalone mode (a PyInstaller module), we return the absolute path of the directory where the executable is.
    - In development mode, we return the absolute path to the directory where the angr management package is.

    :return:    A string that represents the path to the application that can be used to run angr management.
    """

    if is_pyinstaller():
        # running as a PyInstaller bundle
        return os.path.dirname(sys.executable)
    else:
        # running as a Python package
        return os.path.normpath(os.path.join(os.path.abspath(__file__), "..", ".."))



# pylint:disable=missing-class-docstring
from __future__ import annotations

import unittest

from common import AngrManagementTestCase
from PySide6.QtCore import QModelIndex, Qt
from PySide6.QtGui import QColor, QFont
from PySide6.QtTest import QTest
from PySide6.QtWidgets import QApplication, QComboBox

from angrmanagement.ui.widgets.qproperty_editor import (
    BoolPropertyItem,
    ColorPropertyItem,
    ComboPropertyItem,
    FilePropertyItem,
    FloatPropertyItem,
    FontPropertyItem,
    GroupPropertyItem,
    IntPropertyItem,
    PropertyModel,
    QPropertyEditor,
    TextPropertyItem,
    for_all_model_rows,
)

SLOW_TEST = False


def w():
    if SLOW_TEST:
        QTest.qWait(1000)


class TestQPropertyEditor(AngrManagementTestCase):
    def setUp(self):
        super().setUp()
        self.root = GroupPropertyItem(
            "Root",
            description="Root of all properties.",
            children=[
                GroupPropertyItem(
                    "General",
                    description="General settings for the object.",
                    children=[
                        TextPropertyItem("Name", "Sample Object", description="The name of the object."),
                        BoolPropertyItem(
                            "Enabled",
                            True,
                            description="Enable or disable the object.",
                            children=[
                                TextPropertyItem("Subprop1", "Sub-property 1", description="I'm a sub-property"),
                                TextPropertyItem(
                                    "Subprop2",
                                    "Sub-property 2",
                                    description="I'm a sub-property",
                                    children=[
                                        TextPropertyItem(
                                            "Subsubprop", "Sub-sub-property", description="I'm a sub-sub-property"
                                        ),
                                    ],
                                ),
                            ],
                        ),
                        FontPropertyItem("Font", QFont("Arial", 10), description="Select a font."),
                        FilePropertyItem("File", "", description="Select a file."),
                    ],
                ),
                GroupPropertyItem(
                    "Size",
                    description="Size settings.",
                    children=[
                        IntPropertyItem("Width", 100, minimum=0, maximum=1000, description="The width in pixels."),
                        IntPropertyItem("Height", 200, minimum=0, maximum=1000, description="The height in pixels."),
                    ],
                ),
                GroupPropertyItem(
                    "Appearance",
                    description="Visual appearance settings.",
                    children=[
                        FloatPropertyItem(
                            "Opacity",
                            0.85,
                            minimum=0.0,
                            maximum=1.0,
                            decimals=2,
                            description="The opacity (transparency level).",
                        ),
                        ColorPropertyItem("Background Color", QColor("blue"), description="The background color."),
                        ComboPropertyItem(
                            "Style", "Solid", ["Solid", "Dashed", "Dotted"], description="The border style."
                        ),
                    ],
                ),
            ],
        )

        self.model = PropertyModel(self.root)
        self.tree_view = QPropertyEditor()
        self.tree_view.setModel(self.model)
        self.tree_view.setWindowTitle("Property Editor")
        self.tree_view.resize(500, 500)
        self.tree_view.show()

        self.tv = self.tree_view._tree_view
        self.tv.setFocus()

        w()

    def tearDown(self):
        super().tearDown()
        self.tree_view.hide()
        del self.tree_view

    def test_desc_box_updates(self):
        general_prop = self.root.children[0]
        self._click_on_prop(general_prop)
        assert general_prop.name in self.tree_view._desc_box.toPlainText()
        w()

        name_prop = self.root.children[0].children[0]
        self._click_on_prop(name_prop)
        assert name_prop.description in self.tree_view._desc_box.toPlainText()
        w()

    def test_text_prop(self):
        prop = self.root.children[0].children[0]
        assert isinstance(prop, TextPropertyItem)

        # Double click to edit
        self._click_on_prop(prop, double=True)
        w()

        # Update text
        new_text = "New Value"
        QTest.keyClicks(self.tv.focusWidget(), new_text)
        QTest.keyClick(self.tv.focusWidget(), Qt.Key.Key_Return)
        QApplication.processEvents()
        w()

        self.assertEqual(prop.value, new_text)

    def test_bool_prop(self):
        prop = self.root.children[0].children[1]
        assert isinstance(prop, BoolPropertyItem)

        initial_value = prop.value

        # Click on it
        self._click_on_prop(prop)
        QApplication.processEvents()
        assert prop.value == (not initial_value)
        w()

        # Click again
        self._click_on_prop(prop)
        QApplication.processEvents()
        assert prop.value == initial_value
        w()

    def test_combo_prop(self):
        prop = self.root.children[2].children[2]
        assert isinstance(prop, ComboPropertyItem)

        initial_value = prop.value

        # Double click to edit
        self._click_on_prop(prop, double=True)
        w()

        # Select next item
        combo = self.tv.focusWidget()
        assert isinstance(combo, QComboBox)

        combo.setCurrentIndex(combo.currentIndex() + 1)
        new_data = combo.currentData()
        assert new_data != initial_value
        w()

        # Accept
        QTest.keyClick(combo, Qt.Key.Key_Return)
        QApplication.processEvents()
        assert prop.value == new_data
        w()

    def test_expand_prop_with_double_click(self):
        for prop in self._iter_props(self.root):
            if not prop.children:
                continue

            idx = self._prop_to_index(prop)

            # Collapse
            self._click_on_prop(prop, col=0, double=True)
            assert not self.tv.isExpanded(idx)
            w()

            # Expand
            self._click_on_prop(prop, col=0, double=True)
            assert self.tv.isExpanded(idx)
            w()

    def test_expand_prop_with_arrow(self):
        for prop in self._iter_props(self.root):
            if not prop.children:
                continue

            idx = self._prop_to_index(prop)

            # Get branch arrow widget location
            vr = self.tv.visualRect(idx)
            pos = vr.center()
            pos.setX(int(vr.height() / 2))

            # Collapse
            QTest.mouseClick(self.tv.viewport(), Qt.MouseButton.LeftButton, pos=pos)
            assert not self.tv.isExpanded(idx)
            w()

            # Expand
            QTest.mouseClick(self.tv.viewport(), Qt.MouseButton.LeftButton, pos=pos)
            assert self.tv.isExpanded(idx)
            w()

    def test_filter(self):
        text = "subprop1"
        prop = self.root.children[0].children[1].children[0]

        # Gather expected (all parents but root)
        expected_visible = {prop}
        while prop.parent and prop.parent.parent:
            expected_visible.add(prop.parent)
            prop = prop.parent

        # Gather initially visible
        initially_visible = self._get_visible_props()

        # Set filter text
        self.tree_view._filter_box.setText(text)
        w()

        # Check visible items
        assert expected_visible == self._get_visible_props()
        w()

        # Clear filter
        self.tree_view._filter_box.setText("")
        assert initially_visible == self._get_visible_props()
        w()

    # FIXME: Test Color
    # FIXME: Test Int
    # FIXME: Test Float
    # FIXME: Test File
    # FIXME: Test Font

    def _iter_props(self, item):
        if item is not self.root:
            yield item
        for c in item.children:
            yield from self._iter_props(c)

    def _prop_to_index(self, prop, col=0) -> QModelIndex:
        indices = []
        for_all_model_rows(self.tree_view._proxy_model, indices.append)
        for index in indices:
            if not index.isValid():
                continue
            prop_at_index = self.tree_view._proxy_model.mapToSource(index).internalPointer()
            if prop_at_index is prop:
                if col:
                    return index.siblingAtColumn(col)
                return index
        raise KeyError("Could not resolve index")

    def _click_on_prop(self, prop, col=1, double=False):
        QApplication.processEvents()
        index = self._prop_to_index(prop, col)
        pos = self.tv.visualRect(index).center()
        vp = self.tv.viewport()

        QApplication.processEvents()

        if double:
            # Bug in QTreeView::mouseDoubleClickEvent? Events are being sent in but the double click doesn't trigger
            QTest.mouseClick(vp, Qt.MouseButton.LeftButton, pos=pos)
            QTest.qWait(250)

            QTest.mouseDClick(vp, Qt.MouseButton.LeftButton, pos=pos)
        else:
            QTest.mouseClick(vp, Qt.MouseButton.LeftButton, pos=pos)

        QApplication.processEvents()

    def _get_visible_props(self):
        visible = set()

        def on_index(index):
            if index.isValid():
                visible.add(self.tree_view._proxy_model.mapToSource(index).internalPointer())

        for_all_model_rows(self.tree_view._proxy_model, on_index)
        return visible


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import networkx

from .edge import Edge


class Direction:
    TOP = 0
    BOTTOM = 1
    BOTH = 2


class TreeGraphEdgeRouter:
    def __init__(
        self,
        layers: list[list[Any]],
        vertical: bool,
        layer_widths: list[float],  # used if vertical is False
        layer_heights: list[float],  # used if vertical is True
        graph: networkx.DiGraph,
        node_coordinates: dict[Any, tuple[float, float]],
        node_sizes: dict[Any, tuple[float, float]],
        horizontal_spacing: int,
        vertical_spacing: int,
    ) -> None:
        self.layers = layers
        self.vertical = vertical
        self.layer_widths = layer_widths
        self.layer_heights = layer_heights
        self.graph = graph
        self.node_coordinates = node_coordinates
        self.node_sizes = node_sizes
        self.horizontal_spacing = horizontal_spacing
        self.vertical_spacing = vertical_spacing

        self.edges = []

        if self.vertical:
            self._route_vertical()
        else:
            self._route_horizontal()

    def _route_vertical(self) -> None:
        curr_y = 0
        for i, layer in enumerate(self.layers):
            curr_y += self.layer_heights[i] + self.vertical_spacing
            for src in layer:
                x0, y0 = self.node_coordinates[src]
                src_width, src_height = self.node_sizes[src]
                for dst in self.graph.successors(src):
                    x1, y1 = self.node_coordinates[dst]
                    # TODO: Right now there is no going back. This should be fixed to handle loops
                    if y0 >= y1:
                        continue
                    dst_width = self.node_sizes[dst][0]
                    edge = Edge(src, dst)
                    x_start = x0 + src_width / 2
                    x_start += 3  # half of the width of the arrow
                    y_start = y0 + src_height
                    x_end = x1 + dst_width / 2
                    y_end = y1 - 6  # height of the arrow
                    if x_start == x_end:
                        # just draw a straight line
                        edge.add_coordinate(x_start, y_start)
                        edge.add_coordinate(x1, y_end)
                    else:
                        # segment line
                        edge.add_coordinate(x_start, y_start)
                        y = curr_y - self.vertical_spacing / 2
                        edge.add_coordinate(x_start, y)
                        edge.add_coordinate(x_end, y)
                        edge.add_coordinate(x_end, y_end)

                    self.edges.append(edge)

    def _route_horizontal(self) -> None:
        curr_x = 0
        for i, layer in enumerate(self.layers):
            curr_x += self.layer_widths[i] + self.horizontal_spacing
            for src in layer:
                x0, y0 = self.node_coordinates[src]
                src_width, src_height = self.node_sizes[src]
                for dst in self.graph.predecessors(src):
                    x1, y1 = self.node_coordinates[dst]
                    # TODO: Right now there is no going back. This should be fixed to handle loops
                    if x0 >= x1:
                        continue
                    dst_height = self.node_sizes[dst][1]
                    edge = Edge(src, dst)
                    x_start = x0 + src_width
                    x_start += 6  # width of the arrow
                    y_start = y0 + src_height / 2
                    y_end = y1 + dst_height / 2
                    if y_start == y_end:
                        # just draw a straight line
                        edge.add_coordinate(x_start, y_start)
                        edge.add_coordinate(x1, y_end)
                    else:
                        # segment line
                        edge.add_coordinate(x_start, y_start)
                        x = curr_x - self.horizontal_spacing / 2
                        edge.add_coordinate(x, y_start)
                        edge.add_coordinate(x, y_end)
                        x = x1
                        edge.add_coordinate(x, y_end)

                    self.edges.append(edge)


class TreeGraphLayouter:
    """
    Vertically or horizontally layout a tree-like graph.
    """

    def __init__(
        self,
        graph: networkx.DiGraph,
        node_sizes: dict[Any, tuple[float, float]],
        initial_nodes: list[Any] | None = None,
        vertical: bool = False,
        direction: int = Direction.BOTH,
        top_limit: int | None = None,
        bottom_limit: int | None = None,
        horizontal_spacing: int = 145,
        vertical_spacing: int = 15,
        layer_sorter=None,
    ) -> None:
        self._graph = graph
        self._node_sizes = node_sizes

        self._initial_nodes = initial_nodes
        self._top_limit = top_limit
        self._bottom_limit = bottom_limit
        self._direction = direction
        self._vertical = vertical
        self.horizontal_spacing = horizontal_spacing
        self.vertical_spacing = vertical_spacing
        self._layer_sorter = layer_sorter

        self.node_coordinates: dict[Any, tuple[float, float]] = {}
        self.edges = []

        self._layout()

    def _layout(self) -> None:
        layers: list[list[Any]] = []

        if not self._initial_nodes:
            # use root nodes as the initial nodes
            # assuming root nodes are not within any loop
            initial_nodes = [n for n in self._graph.nodes() if self._graph.out_degree[n] == 0]
        else:
            initial_nodes = self._initial_nodes

        layers.append(initial_nodes)
        existing_nodes: set[Any] = set(initial_nodes)

        if self._direction in (Direction.BOTTOM, Direction.BOTH):
            # expand to include successors
            i = 0
            last_layer = layers[0]
            while self._bottom_limit is None or i < self._bottom_limit:
                i += 1
                new_layer = []
                for node in last_layer:
                    for succ in self._graph.successors(node):
                        if succ not in existing_nodes:
                            new_layer.append(succ)
                            existing_nodes.add(succ)
                if not new_layer:
                    break
                if self._layer_sorter:
                    new_layer = self._layer_sorter(new_layer)
                layers.insert(0, new_layer)
                last_layer = new_layer

        if self._direction in (Direction.TOP, Direction.BOTH):
            # expand to include predecessors
            i = 0
            last_layer = layers[-1]
            while self._top_limit is None or i < self._top_limit:
                i += 1
                new_layer = []
                for node in last_layer:
                    for pred in self._graph.predecessors(node):
                        if pred not in existing_nodes:
                            new_layer.append(pred)
                            existing_nodes.add(pred)
                if not new_layer:
                    break
                if self._layer_sorter:
                    new_layer = self._layer_sorter(new_layer)
                layers.append(new_layer)
                last_layer = new_layer

        # layout each layer, from root nodes to leaves
        layer_widths = []
        layer_heights = []
        x, y = 0.0, 0.0

        if self._vertical:
            # Tree layers are vertically layouted
            layers = layers[::-1]
            for layer in layers:
                layer_width, layer_height = self._layout_layer_vertical(x, y, layer)
                y += layer_height + self.vertical_spacing
                x = 0.0
                layer_heights.append(layer_height)
        else:
            # Tree layers are horizontally layouted
            for layer in layers:
                layer_width, layer_height = self._layout_layer_horizontal(x, y, layer)
                x += layer_width + self.horizontal_spacing
                y = 0.0
                layer_widths.append(layer_width)

        # edges
        self.edges = TreeGraphEdgeRouter(
            layers,
            self._vertical,
            layer_widths,
            layer_heights,
            self._graph,
            self.node_coordinates,
            self._node_sizes,
            self.horizontal_spacing,
            self.vertical_spacing,
        ).edges

    def _max_width_and_height(self, nodes) -> tuple[float, float]:
        max_width, max_height = 0.0, 0.0

        # calculate max width and max height
        for node in nodes:
            width_, height_ = self._node_sizes[node]
            if width_ > max_width:
                max_width = width_
            if height_ > max_height:
                max_height = height_

        return max_width, max_height

    def _layout_layer_vertical(self, x, y, nodes) -> tuple[float, float]:
        """
        Vertically layout a layer of nodes.
        """
        max_width, max_height = self._max_width_and_height(nodes)

        # calculate their coordinates
        curr_x = x

        for node in nodes:
            preds = self._graph.predecessors(node)
            min_x = None
            max_x = None
            for pred in preds:
                if pred in self.node_coordinates:
                    pred_x, _ = self.node_coordinates[pred]
                    if min_x is None or pred_x < min_x:
                        min_x = pred_x
                    if max_x is None or pred_x + self._node_sizes[pred][0] > max_x:
                        max_x = pred_x + pred.width

            width_, height_ = self._node_sizes[node]
            y_ = y + (max_height / 2 - height_ / 2)

            if min_x is None or max_x is None:
                # preds don't exist
                # just give it something
                x_ = curr_x
            else:
                # stay in the middle of the preds
                x_ = min_x + (max_x - min_x) / 2 - width_ / 2
                if x_ < curr_x:
                    x_ = curr_x

            self.node_coordinates[node] = (x_, y_)
            curr_x = x_ + width_ + self.horizontal_spacing

        return curr_x, max_height

    def _layout_layer_horizontal(self, x, y, nodes) -> tuple[float, float]:
        """
        Horizontally layout a layer of nodes.
        """
        max_width, max_height = self._max_width_and_height(nodes)

        # calculate their coordinates
        curr_y = y

        for node in nodes:
            preds = self._graph.predecessors(node)
            min_y = None
            max_y = None
            for pred in preds:
                if pred in self.node_coordinates:
                    _, pred_y = self.node_coordinates[pred]
                    if min_y is None or pred_y < min_y:
                        min_y = pred_y
                    if max_y is None or pred_y + self._node_sizes[pred][1] > max_y:
                        max_y = pred_y + pred.height

            width_, height_ = self._node_sizes[node]
            x_ = x + (max_width / 2 - width_ / 2)

            if min_y is None or max_y is None:
                # preds don't exist
                # just give it something
                y_ = curr_y
            else:
                # stay in the middle of the preds
                y_ = min_y + (max_y - min_y) / 2 - height_ / 2
                if y_ < curr_y:
                    y_ = curr_y

            self.node_coordinates[node] = (x_, y_)
            curr_y = y_ + height_ + self.vertical_spacing

        return max_width, curr_y



from __future__ import annotations

from PySide6.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout, QLabel, QLineEdit, QPushButton, QVBoxLayout


class LabelNameBox(QLineEdit):
    """
    A QLineEdit that sanitizes label names.
    """

    def __init__(self, textchanged_callback, parent=None) -> None:
        super().__init__(parent)

        self.textChanged.connect(textchanged_callback)

    @property
    def label(self):
        text = self.text()
        if not text or self._is_valid_label_name(text):
            return text.strip()
        return None

    def _is_valid_label_name(self, input_):  # pylint: disable=no-self-use
        return input_ and " " not in input_.strip()


class RenameLabel(QDialog):
    """
    Dialog to rename labels.
    """

    def __init__(self, disasm_view, label_addr, full_refresh: bool = False, parent=None) -> None:
        super().__init__(parent)

        # initialization
        self._disasm_view = disasm_view
        self._label_addr = label_addr
        self._full_refresh = full_refresh
        self._label_type: str = None

        self._name_box: LabelNameBox = None
        self._status_label = None
        self._ok_button: QPushButton = None

        self.setWindowTitle(f"Rename Label at {self._label_addr:#08x}")

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        # name label

        name_label = QLabel(self)
        name_label.setText("New name")

        name_box = LabelNameBox(self._on_name_changed, self)
        text = ""
        if self._disasm_view.disasm.kb.functions.contains_addr(self._label_addr):
            self._label_type = "function"
            text = self._disasm_view.disasm.kb.functions.get_by_addr(self._label_addr).name
        else:
            self._label_type = "label"
            if self._label_addr in self._disasm_view.disasm.kb.labels:
                text = self._disasm_view.disasm.kb.labels[self._label_addr]

        if text:
            name_box.setText(text)
            name_box.selectAll()
        self._name_box = name_box

        label_layout = QHBoxLayout()
        label_layout.addWidget(name_label)
        label_layout.addWidget(name_box)
        self.main_layout.addLayout(label_layout)

        # status label
        status_label = QLabel(self)
        self.main_layout.addWidget(status_label)
        self._status_label = status_label

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self._ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_name_changed(self, new_text) -> None:  # pylint: disable=unused-argument
        if self._name_box is None:
            # initialization is not done yet
            return

        if self._name_box.label is None:
            # the label name is invalid

            self._status_label.setText("Invalid")
            self._status_label.setProperty("class", "status_invalid")
            self._ok_button.setEnabled(False)
        else:
            self._status_label.setText("Valid")
            self._status_label.setProperty("class", "status_valid")
            self._ok_button.setEnabled(True)

        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)

    def _on_ok_clicked(self) -> None:
        label = self._name_box.label
        if label is not None:
            self._disasm_view.rename_label(
                self._label_addr, label, is_func=self._label_type == "function", full_refresh=self._full_refresh
            )
            self.close()



from __future__ import annotations


class SynchronizedHighlightRegion:
    """
    A region of memory to be highlighted in synchronized views.
    """

    def __init__(self, addr: int, size: int) -> None:
        self.addr: int = addr
        self.size: int = size



from __future__ import annotations

import logging
import sys
import time
from typing import TYPE_CHECKING

from PySide6.QtCore import QObject, QThread, Signal, Slot
from PySide6.QtWidgets import QApplication

from angrmanagement.data.jobs.job import Job, JobState

if TYPE_CHECKING:
    from collections.abc import Callable

    from angrmanagement.ui.workspace import Workspace


log = logging.getLogger(__name__)


class JobCancelled(BaseException):
    """Raised when a job is cancelled."""


class JobContext:
    """Passed to each job to allow it to report progress."""

    _job: Job
    _job_worker: Worker
    _last_reported_timestamp: float
    _last_reported_percentage: float
    _last_text: str

    def __init__(self, worker: Worker, job: Job):
        self._job_worker = worker
        self._job = job
        self._last_reported_timestamp = 0.0
        self._last_reported_percentage = 0.0
        self._last_text = ""

    def set_progress(self, percentage: float, text: str = "") -> None:
        if self._job.state == JobState.CANCELLED:
            raise JobCancelled

        self._job.progress_percentage = percentage

        percentage_delta = percentage - self._last_reported_percentage
        time_delta = time.time() - self._last_reported_timestamp

        if (percentage_delta > 0.02 or self._last_text != text) and time_delta >= 0.1:
            self._last_reported_percentage = percentage
            self._last_reported_timestamp = time.time()
            self._job_worker.job_progressed.emit(self._job, percentage, text)


class Worker(QObject):
    """Executes jobs."""

    job_starting = Signal(Job)
    job_progressed = Signal(Job, float, str)
    job_exception = Signal(Job, BaseException)
    job_finished = Signal(Job)

    def __init__(self, job_manager: JobManager):
        super().__init__()
        self.job_manager = job_manager

    @Slot(Job)
    def execute_job(self, job: Job) -> None:
        if job.state != JobState.PENDING:
            return

        self.job_starting.emit(job)

        try:
            job.state = JobState.RUNNING
            job.start_at = time.time()
            log.info('Job "%s" started', job.name)
            ctx = JobContext(self, job)
            ctx.set_progress(0.0)
            job.start(ctx)
            duration = time.time() - job.start_at
            log.info('Job "%s" completed after %.2f seconds', job.name, duration)

            if job.state == JobState.RUNNING:
                job.state = JobState.FINISHED
                ctx.set_progress(100.0)
            self.job_finished.emit(job)

        except JobCancelled as e:
            log.exception("Job successfully cancelled")
            self.job_exception.emit(job, e)

        except Exception as e:  # pylint: disable=broad-except
            log.exception('Exception while running job "%s":', job.name)
            job.state = JobState.FAILED
            self.job_exception.emit(job, e)


class JobManager(QObject):
    """Manages job execution in a worker thread."""

    workspace: Workspace
    jobs: list[Job]
    job_worker_exception_callback: Callable[[Job, BaseException], None] | None

    job_added = Signal(Job)
    job_starting = Signal(Job)
    job_progressed = Signal(Job, float, str)
    job_exception = Signal(Job, BaseException)
    job_finished = Signal(Job)

    _worker_thread: QThread
    _worker: Worker

    def __init__(self, workspace: Workspace):
        super().__init__()

        self.workspace = workspace
        self.jobs = []
        self._current_job: Job | None = None
        self.job_worker_exception_callback = None

        self._worker_thread = QThread()
        self._worker = Worker(self)
        self._worker.moveToThread(self._worker_thread)
        self._worker.job_starting.connect(self._on_job_starting)
        self._worker.job_progressed.connect(self._on_job_progress)
        self._worker.job_exception.connect(self._on_job_exception)
        self._worker.job_finished.connect(self._on_job_finished)
        self.job_added.connect(self._worker.execute_job)
        self._worker_thread.start()

    def quit(self):
        self._worker_thread.quit()
        self._worker_thread.wait()

    def add_job(self, job: Job) -> None:
        self.jobs.append(job)
        self.job_added.emit(job)

    @staticmethod
    def cancel_job(job: Job) -> None:
        """
        Cancel a job.
        """
        job.state = JobState.CANCELLED

    def interrupt_current_job(self) -> None:
        """Notify the current running job that the user requested an interrupt. The job may ignore it."""
        if self._current_job:
            self._current_job.state = JobState.CANCELLED

    def join_all_jobs(self, wait_period: float = 2.0) -> None:
        """
        Wait until self.jobs is empty for at least `wait_period` seconds.

        This is because one job may add another job upon completion. We cannot simply wait until self.jobs becomes
        empty.
        """
        last_has_job = time.time()
        while time.time() - last_has_job <= wait_period:
            while self.jobs:
                QApplication.processEvents()
                last_has_job = time.time()
                time.sleep(0.05)

    def _on_job_starting(self, job):
        self._current_job = job
        self.job_starting.emit(job)

    def _on_job_progress(self, job: Job, percentage: float, text: str = "") -> None:
        self.job_progressed.emit(job, percentage, text)

    def _on_job_exception(self, job, e):
        if self.job_worker_exception_callback is not None:
            self.job_worker_exception_callback(job, e)

        if job in self.jobs:
            self.jobs.remove(job)

        # Store exception for console debugging
        sys.last_traceback = e.__traceback__
        sys.last_value = e
        sys.last_type = type(e)
        sys.last_exc = e

        self._current_job = None
        self.job_exception.emit(job, e)

    def _on_job_finished(self, job):
        job.finish()  # Job finished handler (GUI thread only)

        if job in self.jobs:
            self.jobs.remove(job)

        self._current_job = None
        self.job_finished.emit(job)



from __future__ import annotations

from PySide6.QtGui import QColor
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
)

from .function_diff import BFSFunctionDiff, LinearFunctionDiff


class SettingsDialog(QDialog):
    """
    A settings dialog for the Precise Diff plugin
    """

    def __init__(self, diff_plugin, parent=None) -> None:
        super().__init__(parent)
        self.setWindowTitle("Precise Diff Settings")
        self.diff_plugin = diff_plugin

        self._main_layout = QVBoxLayout()
        self._init_widgets()
        self.setLayout(self._main_layout)
        self.updates = False
        self.show()

    def _init_widgets(self) -> None:
        upper_layout = QVBoxLayout()

        #
        # Algorithm Choices
        #

        algo_group = QGroupBox(self)
        algo_group.setTitle("Precise Diffing Algorithm")
        algo_group_layout = QVBoxLayout()
        self._bfs_diff_btn = QRadioButton("Graph Breadth First Search")
        self._bfs_diff_btn.setToolTip(
            """
            Diffs two functions by traversing the graph in a BFS manner.
            Blocks that don't exist, or are out of order, in the new graph will be marked
            as ADDed instructions.
            """
        )
        self._linear_diff_btn = QRadioButton("Address Linear")
        self._linear_diff_btn.setToolTip(
            """
            Diffs two functions by linearly traversing the assembly of the original
            and checking for the same index in the new binary.
            """
        )
        self._use_addrs = QCheckBox("Use addresses for alignment")
        self._use_addrs.setToolTip(
            """
            When enabled, the diffing algorithm will attempt to diff functions at the same addresses across
            both binaries. When disabled we attempt to use symbols.
            """
        )
        self._use_addrs.setChecked(self.diff_plugin.use_addrs)

        if self.diff_plugin.diff_algo_class == BFSFunctionDiff:
            self._bfs_diff_btn.setChecked(True)
            self._linear_diff_btn.setChecked(False)
        elif self.diff_plugin.diff_algo_class == LinearFunctionDiff:
            self._bfs_diff_btn.setChecked(False)
            self._linear_diff_btn.setChecked(True)

        algo_group_layout.addWidget(self._bfs_diff_btn)
        algo_group_layout.addWidget(self._linear_diff_btn)
        algo_group_layout.addWidget(self._use_addrs)
        algo_group.setLayout(algo_group_layout)

        upper_layout.addWidget(algo_group)

        #
        # Instruction
        #

        ins_group = QGroupBox(self)
        ins_group.setTitle("Instruction Diffing Options")
        ins_layout = QVBoxLayout()
        self._prefer_symbols = QCheckBox("Prioritize Symbols", self)
        self._prefer_symbols.setToolTip(
            """
            Some instructions that use symbols, such as moves from global vars, may show different addresses
            in the new binary. With this option enabled, two instructions are marked as the same if at least
            the symbol name lines up.
            """
        )
        self._prefer_symbols.setChecked(self.diff_plugin.prefer_symbols)
        self._prefer_strings = QCheckBox("Prioritize Strings", self)
        self._prefer_strings.setToolTip(
            """
            Some instructions that use strings, such as moves from memory, may show different addresses
            in the new binary. With this option enabled, two instructions that move an address that point to the
            same string are marked as the same.
            """
        )
        self._prefer_strings.setChecked(self.diff_plugin.resolve_strings)
        self._prefer_insns = QCheckBox("Prioritize Instructions", self)
        self._prefer_insns.setToolTip(
            """
            Some instructions that use addresses, such as jumps, may show different addresses
            in the new binary. With this option enabled, two instructions that use an address that both point
            to the same first few instructions are marked as the same.
            """
        )
        self._prefer_insns.setChecked(self.diff_plugin.resolve_insns)
        self._ignore_globals = QCheckBox("Ignore Unnamed Globals")
        self._ignore_globals.setToolTip(
            """
            When enabled, ignores all unnamed global values when showing psuedocode diffs.
            """
        )
        self._ignore_globals.setChecked(self.diff_plugin.ignore_globals)

        ins_layout.addWidget(self._prefer_symbols)
        ins_layout.addWidget(self._prefer_strings)
        ins_layout.addWidget(self._prefer_insns)
        ins_layout.addWidget(self._ignore_globals)
        ins_group.setLayout(ins_layout)

        upper_layout.addWidget(ins_group)

        #
        # GUI Settings
        #

        gui_group = QGroupBox(self)
        gui_group.setTitle("GUI Options")
        gui_layout = QGridLayout()

        change_label = QLabel("Diff Change Color", self)
        change_label.setToolTip("The color shown when two instructions differ in some sub-change (like ops)")
        change_label.setStyleSheet(f"background-color: #{self.diff_plugin.chg_color.rgba() & 0xffffff:x}")
        self._change_color = QLineEdit()
        self._change_color.setText(hex(self.diff_plugin.chg_color.rgb()))
        gui_layout.addWidget(change_label, 0, 0)
        gui_layout.addWidget(self._change_color, 0, 1)

        add_label = QLabel("Diff Add Color", self)
        add_label.setToolTip("The color shown when an instruction can't be matched so it's assumed to be new")
        add_label.setStyleSheet(f"background-color: #{self.diff_plugin.add_color.rgba() & 0xffffff:x}")
        self._add_color = QLineEdit()
        self._add_color.setText(hex(self.diff_plugin.add_color.rgb()))
        gui_layout.addWidget(add_label, 1, 0)
        gui_layout.addWidget(self._add_color, 1, 1)

        del_label = QLabel("Diff Delete Color", self)
        del_label.setToolTip(
            "The color shown in the original binary where an instruction was removed (disabled for now)"
        )
        del_label.setStyleSheet(f"background-color: #{self.diff_plugin.del_color.rgba() & 0xffffff:x}")
        self._del_color = QLineEdit()
        self._del_color.setText(hex(self.diff_plugin.del_color.rgb()))
        gui_layout.addWidget(del_label, 2, 0)
        gui_layout.addWidget(self._del_color, 2, 1)
        gui_group.setLayout(gui_layout)

        upper_layout.addWidget(gui_group)

        #
        # Ok and Cancel Buttons
        #

        self._ok_button = QPushButton(self)
        self._ok_button.setText("OK")
        self._ok_button.setDefault(True)
        self._ok_button.clicked.connect(self._on_ok_clicked)

        cancel_button = QPushButton(self)
        cancel_button.setText("Cancel")
        cancel_button.clicked.connect(self._on_cancel_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(cancel_button)

        # main layout
        self._main_layout.addLayout(upper_layout)
        self._main_layout.addLayout(buttons_layout)

    def _on_ok_clicked(self) -> None:
        # algorithms
        if self._bfs_diff_btn.isChecked():
            self.diff_plugin.diff_algo_class = BFSFunctionDiff
        else:
            self.diff_plugin.diff_algo_class = LinearFunctionDiff
        self.diff_plugin.use_addrs = self._use_addrs.isChecked()

        # instruction options
        self.diff_plugin.prefer_symbols = self._prefer_symbols.isChecked()
        self.diff_plugin.resolve_strings = self._prefer_strings.isChecked()
        self.diff_plugin.resolve_insns = self._prefer_insns.isChecked()

        # colors
        try:
            self.diff_plugin.add_color = QColor(int(self._add_color.text(), 16))
            self.diff_plugin.chg_color = QColor(int(self._change_color.text(), 16))
            self.diff_plugin.del_color = QColor(int(self._del_color.text(), 16))
        except ValueError:
            pass

        self.updates = True
        self.close()

    def _on_cancel_clicked(self) -> None:
        self.close()



# pylint:disable=unused-private-member
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

_l = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterable

    from angr.analyses.decompiler.optimization_passes.optimization_pass import OptimizationPass
    from angr.sim_manager import SimulationManager
    from PySide6.QtGui import QColor, QIcon, QPainter
    from PySide6.QtWidgets import QGraphicsSceneMouseEvent

    from angrmanagement.config.config_entry import ConfigurationEntry
    from angrmanagement.ui.views import code_view, disassembly_view
    from angrmanagement.ui.widgets.qblock import QBlock
    from angrmanagement.ui.widgets.qinst_annotation import QInstructionAnnotation
    from angrmanagement.ui.widgets.qinstruction import QInstruction
    from angrmanagement.ui.workspace import Workspace

# pylint: disable=no-self-use,unused-argument


class BasePlugin:
    """
    Implements the base class for all angr management plugins.
    """

    # Override DISPLAY_NAME to specify a human-readable name for your plugin
    DISPLAY_NAME = None
    REQUIRE_WORKSPACE = True
    __i_hold_this_abstraction_token = True

    def __init__(self, workspace: Workspace) -> None:
        self.workspace: Workspace | None = workspace
        _l.info("Loaded plugin %s", self.__class__.__name__)

        # valid things that we want you do be able to do in __init__:
        # - set callbacks for object containers
        # ... so then all of those should be undone in teardown

    def teardown(self) -> None:
        pass

    @classmethod
    def get_display_name(cls):
        display_name = getattr(cls, "DISPLAY_NAME", None)
        if not display_name:
            return cls.__name__
        return display_name

    #
    # Generic callbacks
    #

    def status_bar_permanent_widgets(self) -> Generator | None:
        """
        Yields all widgets that should be added to the right side of the status bar of the main window.
        """
        return None

    def on_workspace_initialized(self, workspace: Workspace) -> None:
        """
        A handler that is called right after a workspace is initialized.
        """

    def angrdb_store_entries(self):
        """
        Yields all entries (key-value pairs) that should be persisted inside angrDb.
        :return:
        """
        return None

    def angrdb_load_entry(self, key: str, value: str) -> None:
        """
        Called for each entry (key-value pair) that is persisted in angrDb.

        :param key:     Key of the entry.
        :param value:   Value of the entry.
        """

    #
    # UI Callbacks
    #

    def color_insn(self, addr: int, selected, disasm_view) -> QColor | None:
        return None

    def color_block(self, addr: int) -> QColor | None:
        return None

    def color_func(self, func) -> QColor | None:
        return None

    def draw_insn(self, qinsn: QInstruction, painter: QPainter) -> None:
        pass

    def draw_block(self, qblock: QBlock, painter: QPainter) -> None:
        pass

    def instrument_disassembly_view(self, dview: disassembly_view.DisassemblyView) -> None:
        pass

    def instrument_code_view(self, cview: code_view.CodeView) -> None:
        pass

    def handle_click_insn(self, qinsn, event: QGraphicsSceneMouseEvent) -> bool:
        return False

    def handle_click_block(self, qblock, event: QGraphicsSceneMouseEvent) -> bool:
        return False

    def handle_raise_view(self, view) -> None:
        pass

    # iterable of tuples (icon, tooltip)
    TOOLBAR_BUTTONS: list[tuple[QIcon, str]] = []

    def handle_click_toolbar(self, idx: int) -> None:
        pass

    # Iterable of button texts
    MENU_BUTTONS: list[str] = []

    def handle_click_menu(self, idx: int) -> None:
        pass

    # Iterable of column names
    FUNC_COLUMNS: list[str] = []

    def extract_func_column(self, func, idx: int) -> tuple[Any, str]:
        # should return a tuple of the sortable column data and the rendered string
        return 0, ""

    def build_context_menu_insn(self, item) -> Iterable[None | tuple[str, Callable]]:
        """
        Use None to insert a MenuSeparator(). The tuples are: (menu entry text, callback)
        """
        return []

    def build_context_menu_block(self, item) -> Iterable[None | tuple[str, Callable]]:
        """
        Use None to insert a MenuSeparator(). The tuples are: (menu entry text, callback)
        """
        return []

    def build_context_menu_node(self, node) -> Iterable[None | tuple[str, Callable]]:
        """
        Use None to insert a MenuSeparator(). The tuples are: (menu entry text, callback)
        """
        return []

    def build_context_menu_functions(self, funcs) -> Iterable[None | tuple[str, Callable]]:
        return []

    def build_qblock_annotations(self, qblock: QBlock) -> Iterable[QInstructionAnnotation]:
        return []

    # Iterable of URL actions
    URL_ACTIONS: list[str] = []

    def handle_url_action(self, action, kwargs) -> None:
        pass

    def step_callback(self, simgr: SimulationManager) -> None:
        pass

    # Custom configuration entries
    CONFIG_ENTRIES: list[ConfigurationEntry] = []

    #
    # Decompiler Callbacks
    #

    OPTIMIZATION_PASSES: list[tuple[type[OptimizationPass], bool]] = []

    def handle_stack_var_renamed(self, func, offset: int, old_name: str, new_name: str) -> bool:
        return False

    def handle_stack_var_retyped(self, func, offset: int, old_type, new_type) -> bool:
        return False

    def handle_func_arg_renamed(self, func, offset: int, old_name: str, new_name: str) -> bool:
        return False

    def handle_func_arg_retyped(self, func, offset: int, old_type, new_type) -> bool:
        return False

    def handle_global_var_renamed(self, address, old_name: str, new_name: str) -> bool:
        return False

    def handle_global_var_retyped(self, address, old_type, new_type) -> bool:
        return False

    def handle_other_var_renamed(self, var, old_name: str, new_name: str) -> bool:
        return False

    def handle_other_var_retyped(self, var, old_type, new_type) -> bool:
        return False

    def handle_function_renamed(self, func, old_name: str, new_name: str) -> bool:
        return False

    def handle_function_retyped(self, func, old_type, new_type) -> bool:
        return False

    def handle_comment_changed(self, address, old_cmt, new_cmt, created: bool, decomp: bool) -> bool:
        return False

    def handle_struct_changed(self, old_struct, new_struct) -> bool:
        return False

    def decompile_callback(self, func) -> None:
        """
        A callback that is called *right after* the decompiler is run on a function. You can access the current codegen
        with ``self.workspace.instance.kb.decompilations[(func.addr, 'pseudocode')]``
        :param func:        angr Function that was just decompiled
        """

    def handle_project_save(self, file_name: str) -> None:
        """
        A handler to notify plugins whenever the project has been saved by the user.

        @param file_name:       Name in which project is saved as.
        @return:
        """

    def handle_project_initialization(self) -> None:
        """
        A handler to perform any initialization when a new project is loaded
        """



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtGui import QKeySequence

from angrmanagement.ui.icons import icon

from .menu import Menu, MenuEntry, MenuSeparator

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow
    from angrmanagement.ui.toolbars.toolbar import Toolbar


class NewViewMenu(Menu):
    """
    Sub-menu to construct new Views
    """

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__("&New", parent=main_window)

        self.entries.extend(
            [
                MenuEntry(
                    "&Linear Disassembly",
                    main_window.workspace.create_and_show_linear_disassembly_view,
                    icon=icon("disassembly-linear"),
                ),
                MenuEntry(
                    "&Graph Disassembly",
                    main_window.workspace.create_and_show_graph_disassembly_view,
                    shortcut=QKeySequence("Ctrl+N"),
                    icon=icon("disassembly-graph"),
                ),
                MenuSeparator(),
                MenuEntry("&Hex", main_window.workspace.create_and_show_hex_view, icon=icon("hex-view")),
            ]
        )


class ToolbarMenuEntry(MenuEntry):
    """
    Menu item to control toolbar visibility.
    """

    def __init__(self, toolbar_cls: type[Toolbar], main_window: MainWindow) -> None:
        tm = main_window.toolbar_manager
        super().__init__(tm.get_name_for_toolbar_class(toolbar_cls), self.on_toggle, checkable=True)
        self.main_window = main_window
        self.toolbar_cls = toolbar_cls

    @property
    def is_visibile(self) -> bool:
        qtb = self.main_window.toolbar_manager.active.get(self.toolbar_cls, None)
        return qtb is not None and qtb.qtoolbar().isVisible()

    def on_toggle(self) -> None:
        self.main_window.toolbar_manager.set_toolbar_visible_by_class(self.toolbar_cls, not self.is_visibile)

    def update_checked(self) -> None:
        self.checked = self.is_visibile


class ToolbarMenu(Menu):
    """
    Sub-menu to control toolbar visibility.
    """

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__("&Toolbars", parent=main_window)
        tm = main_window.toolbar_manager
        self.entries.extend([MenuEntry("Show all", tm.show_all), MenuEntry("Hide all", tm.hide_all), MenuSeparator()])
        for tb_cls in tm.all_toolbars:
            self.entries.append(ToolbarMenuEntry(tb_cls, main_window))
        self.qmenu().aboutToShow.connect(self.update_checked_entries)

    def update_checked_entries(self) -> None:
        for e in self.entries:
            if isinstance(e, ToolbarMenuEntry):
                e.update_checked()


class ViewMenu(Menu):
    """
    Main View menu
    """

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__("&View", parent=main_window)

        self.entries.extend(
            [
                MenuEntry(
                    "Command Palette...",
                    main_window.show_command_palette,
                    shortcut=QKeySequence("Ctrl+Shift+P"),
                    icon=icon("command-palette"),
                ),
                MenuSeparator(),
                ToolbarMenu(main_window),
                MenuSeparator(),
                MenuEntry("Next Tab", main_window.workspace.view_manager.next_tab, shortcut=QKeySequence("Ctrl+Tab")),
                MenuEntry(
                    "Previous Tab",
                    main_window.workspace.view_manager.previous_tab,
                    shortcut=QKeySequence("Ctrl+Shift+Tab"),
                ),
                MenuSeparator(),
                NewViewMenu(main_window),
                MenuSeparator(),
                MenuEntry(
                    "&Linear Disassembly",
                    main_window.workspace.show_linear_disassembly_view,
                    icon=icon("disassembly-linear"),
                ),
                MenuEntry(
                    "&Graph Disassembly",
                    main_window.workspace.show_graph_disassembly_view,
                    icon=icon("disassembly-graph"),
                ),
                MenuSeparator(),
                MenuEntry("&Hex", main_window.workspace.show_hex_view, icon=icon("hex-view")),
                MenuEntry("Pro&ximity", main_window.view_proximity_for_current_function),
                MenuEntry("Pseudo&code", main_window.workspace.show_pseudocode_view, icon=icon("pseudocode-view")),
                MenuEntry("&Strings", main_window.workspace.show_strings_view, icon=icon("strings-view")),
                MenuEntry("&Patches", main_window.workspace.show_patches_view, icon=icon("patches-view")),
                MenuEntry("&Types", main_window.workspace.show_types_view, icon=icon("types-view")),
                MenuEntry("&Functions", main_window.workspace.show_functions_view, icon=icon("functions-view")),
                MenuEntry("&Traces", main_window.workspace.show_traces_view, icon=icon("traces-view")),
                MenuEntry("&Trace Map", main_window.workspace.show_trace_map_view),
                MenuSeparator(),
                MenuEntry("Symbolic &Execution", main_window.workspace.show_symexec_view),
                MenuEntry("S&ymbolic States", main_window.workspace.show_states_view),
                MenuEntry("&Interaction", main_window.workspace.show_interaction_view),
                MenuEntry("&Registers", main_window.workspace.show_registers_view),
                MenuEntry("&Stack", main_window.workspace.show_stack_view),
                MenuEntry("&Breakpoints", main_window.workspace.show_breakpoints_view),
                MenuEntry("&Call Explorer", main_window.workspace.show_call_explorer_view),
                MenuSeparator(),
                MenuEntry("&Console", main_window.workspace.show_console_view, icon=icon("console-view")),
                MenuEntry("&Log", main_window.workspace.show_log_view, icon=icon("log-view")),
                MenuEntry("&Jobs", main_window.workspace.show_jobs_view, icon=icon("jobs-view")),
            ]
        )



from __future__ import annotations

from .cfg_generation import CFGGenerationJob
from .code_tagging import CodeTaggingJob
from .ddg_generation import DDGGenerationJob
from .decompile_function import DecompileFunctionJob
from .deobfuscation import APIDeobfuscationJob
from .dependency_analysis import DependencyAnalysisJob
from .flirt_signature_recognition import FlirtSignatureRecognitionJob
from .job import Job
from .prototype_finding import PrototypeFindingJob
from .simgr_explore import SimgrExploreJob
from .simgr_step import SimgrStepJob
from .variable_recovery import VariableRecoveryJob
from .vfg_generation import VFGGenerationJob

__all__ = [
    "APIDeobfuscationJob",
    "CFGGenerationJob",
    "CodeTaggingJob",
    "DDGGenerationJob",
    "DecompileFunctionJob",
    "DependencyAnalysisJob",
    "FlirtSignatureRecognitionJob",
    "Job",
    "PrototypeFindingJob",
    "SimgrExploreJob",
    "SimgrStepJob",
    "VariableRecoveryJob",
    "VFGGenerationJob",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QVBoxLayout

from angrmanagement.ui.widgets.qjobs import QJobs

from .view import InstanceView

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class JobsView(InstanceView):
    """JobsView displays all pending, running, and finished jobs in the project."""

    qjobs: QJobs

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("jobs", workspace, default_docking_position, instance)
        self.base_caption = "Jobs"

        # The QJobs widget is initialized in the view, most functions of jobs table is done through QJobs
        self.qjobs = QJobs(workspace)
        vlayout = QVBoxLayout()
        vlayout.addWidget(self.qjobs)
        vlayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(vlayout)
        self.reload()

    def closeEvent(self, event) -> None:
        self.qjobs.close()
        super().closeEvent(event)

    def reload(self) -> None:
        pass

    @staticmethod
    def minimumSizeHint():
        return QSize(0, 50)



from __future__ import annotations

import functools
import os.path

from PySide6.QtGui import QPixmap, Qt
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QFrame,
    QGraphicsScene,
    QGraphicsView,
    QGroupBox,
    QHBoxLayout,
    QPushButton,
    QVBoxLayout,
)

from angrmanagement import __version__
from angrmanagement.config import Conf
from angrmanagement.consts import IMG_LOCATION
from angrmanagement.logic import GlobalInfo
from angrmanagement.ui.icons import icon


class WelcomeDialog(QDialog):
    """
    Welcome dialog.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setWindowTitle("Welcome")
        self._init_widgets()

    def _init_widgets(self) -> None:
        self.setStyleSheet("QPushButton { text-align:left; }")

        # Banner with angr-management version string
        banner_pixmap = QPixmap(os.path.join(IMG_LOCATION, "angr-splash.png"))
        banner_view = QGraphicsView(self)
        banner_view.setContentsMargins(0, 0, 0, 0)
        banner_view.setFrameStyle(QFrame.Shape.NoFrame)
        banner_view.resize(banner_pixmap.size())
        banner_scene = QGraphicsScene()
        banner_view.setScene(banner_scene)
        pi = banner_scene.addPixmap(banner_pixmap)
        font = QApplication.font()
        font.setPointSizeF(9.0)
        ti = banner_scene.addSimpleText(__version__, font)
        ti.setBrush(Qt.GlobalColor.white)
        ti.moveBy(pi.boundingRect().width() - ti.boundingRect().width() - 6, 6)

        # Recent files
        recent_files_group = QGroupBox("Recent Projects")
        recent_files_layout = QVBoxLayout()
        recent_files_group.setLayout(recent_files_layout)

        for path in reversed(Conf.recent_files):
            pb = QPushButton(icon("file"), os.path.basename(path))
            pb.setFlat(True)
            pb.clicked.connect(functools.partial(self._load_recent_file, path))
            recent_files_layout.addWidget(pb)

        recent_files_layout.addStretch()

        # Getting Started actions
        actions_group = QGroupBox("Getting Started")
        actions_layout = QVBoxLayout()

        pb = QPushButton(icon("file-open"), "Open file...")
        pb.setFlat(True)
        pb.clicked.connect(self._open_file)
        actions_layout.addWidget(pb)

        pb = QPushButton(icon("docs"), "Documentation")
        pb.setFlat(True)
        pb.clicked.connect(GlobalInfo.main_window.open_doc_link)
        actions_layout.addWidget(pb)

        pb = QPushButton(icon("about"), "About...")
        pb.setFlat(True)
        pb.clicked.connect(GlobalInfo.main_window.open_about_dialog)
        actions_layout.addWidget(pb)
        actions_layout.addStretch()
        actions_group.setLayout(actions_layout)

        main_layout = QHBoxLayout()
        main_layout.addWidget(recent_files_group)
        main_layout.addWidget(actions_group)
        main_layout.setSpacing(15)
        main_layout.setContentsMargins(15, 15, 15, 15)

        outer_layout = QVBoxLayout()
        outer_layout.setSpacing(0)
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.addWidget(banner_view)
        outer_layout.addLayout(main_layout)
        self.setLayout(outer_layout)

    def _load_recent_file(self, path) -> None:
        GlobalInfo.main_window.load_file(path)
        self.close()

    def _open_file(self) -> None:
        GlobalInfo.main_window.open_file_button()
        self.close()



from __future__ import annotations

from typing import TYPE_CHECKING

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class SimgrExploreJob(InstanceJob):
    """A job that runs the explore method of a simulation manager."""

    def __init__(
        self, instance: Instance, simgr, find=None, avoid=None, step_callback=None, until_callback=None, on_finish=None
    ) -> None:
        super().__init__("Simulation manager exploring", instance, on_finish=on_finish)
        self._simgr = simgr
        self._find = find
        self._avoid = avoid
        self._step_callback = step_callback
        self._until_callback = until_callback
        self._interrupted = False

    def run(self, _: JobContext):
        """Run the job. Runs in the worker thread."""

        def until_callback(*args, **kwargs):
            return self._interrupted or callable(self._until_callback) and self._until_callback(*args, **kwargs)

        self._simgr.explore(find=self._find, avoid=self._avoid, step_func=self._step_callback, until=until_callback)
        return self._simgr

    def __repr__(self) -> str:
        return f"Exploring {self._simgr!r}"

    def cancel(self) -> None:
        """Called from GUI thread. Worker thread will check self._interrupted periodically and exit the job early if
        needed."""
        self._interrupted = True

    @classmethod
    def create(cls, instance: Instance, simgr, **kwargs):
        def callback(result) -> None:
            simgr.am_event(src="job_done", job="explore", result=result)

        return cls(instance, simgr, on_finish=callback, **kwargs)
